# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the functionality in the base class for data model objects"""

# Standard
import importlib
import os

# Third Party
import pytest

# Local
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.data_backends.base import DataModelBackendBase
from caikit.core.data_model.dataobject import DataObjectBase
from tests.data_model_helpers import (
    justify_script_string,
    make_proto_def,
    reset_global_protobuf_registry,
    temp_data_model,
    temp_module,
)

## Helpers #####################################################################


class AccessCounterBackend(DataModelBackendBase):
    """Simple backend that counts how many times a given attribute is accessed
    and can be configured to cache or not.
    """

    def __init__(self, data, do_cache):
        self.data = data
        self.do_cache = do_cache
        self.access_counter = {}

    def cache_attribute(self, name, value):
        return self.do_cache

    # pylint: disable=unused-argument
    def get_attribute(self, data_model_class, name):
        self.access_counter[name] = self.access_counter.get(name, 0) + 1
        return self.data[name]

    def access_count(self, name):
        return self.access_counter.get(name, 0)


@pytest.fixture(autouse=True)
def auto_reset_global_protobuf_registry():
    with reset_global_protobuf_registry():
        yield


## Tests #######################################################################

##########################
## Proto Class Matching ##
##########################


def test_derived_class_no_proto_class():
    """Test that an appropriate error is raised when the protobufs module for a
    derived class does not have a corresponding protobufs class
    """
    with temp_module() as (mod_name, mod_dir):
        # Add a "protobufs.py" file to the library that has nothing in it
        with open(
            os.path.join(mod_dir, "protobufs.py"), "w", encoding="utf-8"
        ) as handle:
            handle.write(justify_script_string("""# Nothing here"""))

        # Add a derived data model class file
        with open(os.path.join(mod_dir, "object.py"), "w", encoding="utf-8") as handle:
            handle.write(
                justify_script_string(
                    """
                    from caikit.core.data_model import base
                    class Object(base.DataBase):
                        def __init__(self, foo):
                            self.foo = foo
                    """
                )
            )

        # Make sure an ValueError is raised when the import is tried
        with pytest.raises(ValueError):
            importlib.import_module(".".join([mod_name, "object"]))


def test_derived_class_no_proto_mod():
    """Test that an appropriate error is raised when the derived library does
    not have a protobufs directory
    """
    with temp_module() as (mod_name, mod_dir):
        # Add a derived data model class file
        with open(os.path.join(mod_dir, "object.py"), "w", encoding="utf-8") as handle:
            handle.write(
                justify_script_string(
                    """
                    from caikit.core.data_model import base
                    class Object(base.DataBase):
                        def __init__(self, foo):
                            self.foo = foo
                    """
                )
            )

        # Make sure an ValueError is raised when the import is tried
        with pytest.raises(ValueError):
            importlib.import_module(".".join([mod_name, "object"]))


def test_derived_class_no_import_side_effects():
    """Test that a library which does not use the import side effects can still
    safely create a derived data model class
    """
    with temp_module() as (mod_name, mod_dir):
        # Add a "protobufs.py" file to the library that has nothing in it
        with open(
            os.path.join(mod_dir, "protobufs.py"), "w", encoding="utf-8"
        ) as handle:
            handle.write(
                justify_script_string(
                    """
                    class FakeDescriptor:
                        fields = []
                        fields_by_name = {}
                        oneofs_by_name = {}
                        oneofs = []
                        name = "Baz"
                        full_name = "foo.bar.Baz"

                    class Object:
                        DESCRIPTOR = FakeDescriptor
                    """
                )
            )

        # Add a derived data model class file
        with open(os.path.join(mod_dir, "object.py"), "w", encoding="utf-8") as handle:
            handle.write(
                justify_script_string(
                    """
                    from . import protobufs
                    from caikit.core.data_model import base
                    class Object(base.DataBase):
                        _proto_class = protobufs.Object
                        def __init__(self, foo):
                            self.foo = foo
                    """
                )
            )

        # Make sure the import succeeds and that the data model class has the
        # right inheritance
        lib = importlib.import_module(".".join([mod_name, "object"]))
        assert hasattr(lib, "Object")
        assert issubclass(lib.Object, DataBase)


def test_compiled_proto_init():
    """Make sure that support for 'compiled' protos works cleanly without using
    the dataobject wrapper
    """
    # pylint: disable=duplicate-code
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": str,
                    "bar": int,
                }
            },
            mock_compiled=True,
        )
    ) as dm:
        assert isinstance(dm.ThingOne, type)
        assert issubclass(dm.ThingOne, DataBase)
        assert not issubclass(dm.ThingOne, DataObjectBase)
        assert dm.ThingOne.fields == ("foo", "bar")

        # Test construction with positional args
        inst = dm.ThingOne("foo", 1)
        assert inst.foo == "foo"
        assert inst.bar == 1

        # Test construction with keyword args
        inst = dm.ThingOne(foo="foo", bar=1)
        assert inst.foo == "foo"
        assert inst.bar == 1

        # Test defaulting to None
        inst = dm.ThingOne()
        assert inst.foo is None
        assert inst.bar is None

        # Test error cases for construction
        with pytest.raises(TypeError):
            dm.ThingOne("foo", foo="bar")
        with pytest.raises(TypeError):
            dm.ThingOne(widget="qewr")
        with pytest.raises(TypeError):
            dm.ThingOne("foo", 1, 2)


def test_compiled_proto_oneof():
    """Make sure that support for 'compiled' protos works cleanly without using
    the dataobject wrapper
    """
    # pylint: disable=duplicate-code
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": "Union[str, int]",
                }
            },
            mock_compiled=True,
        )
    ) as dm:
        assert isinstance(dm.ThingOne, type)
        assert issubclass(dm.ThingOne, DataBase)
        assert not issubclass(dm.ThingOne, DataObjectBase)
        assert set(dm.ThingOne.fields) == {"foo_str", "foo_int"}

        # Construct with the oneof name
        inst = dm.ThingOne(foo=1)
        assert inst.foo == 1
        assert inst.which_oneof("foo") == "foo_int"

        # Construct with field name
        inst = dm.ThingOne(foo_str="asdf")
        assert inst.foo == "asdf"
        assert inst.which_oneof("foo") == "foo_str"

        # Conflicting args
        with pytest.raises(TypeError):
            dm.ThingOne(foo=1, fooster="asdf")


def test_compiled_proto_oneof_union_list():
    """Make sure that support for 'compiled' protos works for Union of lists"""
    # pylint: disable=duplicate-code
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": "Union[List[str], List[int]]",
                }
            },
            mock_compiled=True,
        )
    ) as dm:
        assert isinstance(dm.ThingOne, type)
        assert issubclass(dm.ThingOne, DataBase)
        assert not issubclass(dm.ThingOne, DataObjectBase)
        assert set(dm.ThingOne.fields) == {"foo_int_sequence", "foo_str_sequence"}

        # Construct with the oneof name
        inst = dm.ThingOne(foo=["hello", "world"])
        assert inst.foo == ["hello", "world"]
        assert inst.which_oneof("foo") == "foo_str_sequence"

        # Construct with field name
        inst = dm.ThingOne(foo_int_sequence=[1, 2, 3])
        assert inst.foo == [1, 2, 3]
        assert inst.which_oneof("foo") == "foo_int_sequence"

        # Conflicting args
        with pytest.raises(TypeError):
            dm.ThingOne(foo=["hi"], foo_str_sequence=["asdf"])


##################
## Data Backend ##
##################


def test_cached_backend():
    """Make sure that a data model object will cache the result of get_attribute
    if configured to do so by the backend
    """
    # pylint: disable=duplicate-code
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": str,
                    "bar": int,
                }
            }
        )
    ) as dm:
        data = {"foo": "one", "bar": 2}
        backend = AccessCounterBackend(data, True)
        msg = dm.ThingOne.from_backend(backend)

        # Access the attrs to increment the counters
        msg.foo
        msg.foo
        msg.bar
        msg.bar
        msg.bar
        assert backend.access_count("foo") == 1
        assert backend.access_count("bar") == 1


def test_uncached_backend():
    """Make sure that a data backend can skip the step of caching the data on
    the data model object
    """
    # pylint: disable=duplicate-code
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": str,
                    "bar": int,
                }
            }
        )
    ) as dm:
        data = {"foo": "one", "bar": 2}
        backend = AccessCounterBackend(data, False)
        msg = dm.ThingOne.from_backend(backend)

        # Access the attrs to increment the counters
        msg.foo
        msg.foo
        msg.bar
        msg.bar
        msg.bar
        assert backend.access_count("foo") == 2
        assert backend.access_count("bar") == 3


def test_invalid_attribute_no_backend():
    """Make sure that when created without a backend and without proper
    initialization, an AttributeError is raised
    """
    # pylint: disable=duplicate-code
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": str,
                    "bar": int,
                }
            }
        )
    ) as dm:
        msg = dm.ThingOne.__new__(dm.ThingOne)
        with pytest.raises(AttributeError):
            msg.foo


##################
## Default Init ##
##################


def test_default_init_valid_args():
    """Make sure that valid ways of using the default __init__ all work as
    expected
    """
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": int,
                    "bar": int,
                }
            }
        )
    ) as dm:
        # Both positional
        msg = dm.ThingOne(1, 2)
        assert msg.foo == 1
        assert msg.bar == 2

        # Both kwarg
        msg = dm.ThingOne(foo=1, bar=2)
        assert msg.foo == 1
        assert msg.bar == 2

        # Mix args and kwargs
        msg = dm.ThingOne(1, bar=2)
        assert msg.foo == 1
        assert msg.bar == 2

        # Single arg, other unset
        msg = dm.ThingOne(1)
        assert msg.foo == 1
        assert msg.bar is None

        # Single kwarg, other unset
        msg = dm.ThingOne(foo=1)
        assert msg.foo == 1
        assert msg.bar is None


def test_default_init_invalid_args():
    """Make sure that a TypeError is raised if the default __init__ is given
    various combinations of incorrect arguments
    """
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": int,
                    "bar": int,
                }
            }
        )
    ) as dm:
        # Too many arguments
        with pytest.raises(TypeError):
            dm.ThingOne(1, 2, 3, 4)

        # Too many args + kwargs
        with pytest.raises(TypeError):
            dm.ThingOne(1, 2, baz=3)

        # Bad kwarg name
        with pytest.raises(TypeError):
            dm.ThingOne(1, baz=3)

        # Multiple values
        with pytest.raises(TypeError):
            dm.ThingOne(1, foo=3)


############################
## get_field_message_type ##
############################


def test_get_field_message_type_valid_fields():
    """Make sure that for valid fields, get_field_message_type returns the sub-
    field type on sub-messages and None for other fields
    """
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": int,
                },
                "WrapperThing": {
                    "bar": "ThingOne",
                },
                "RepeatedWrapperThing": {
                    "bar": ["ThingOne"],
                },
            }
        )
    ) as dm:
        # Non-message field
        thing_one = dm.ThingOne(1)
        assert thing_one.get_field_message_type("foo") is None

        # Non-repeated sub-message
        wrapper_msg = dm.WrapperThing(thing_one)
        assert wrapper_msg.get_field_message_type("bar") == dm.ThingOne

        # Repeated sub-message
        dm.RepeatedWrapperThing([thing_one])
        assert wrapper_msg.get_field_message_type("bar") == dm.ThingOne


def test_get_field_message_type_invalid_field():
    """Make sure that an appropriate error is raised for an invalid field"""
    with temp_data_model(
        make_proto_def(
            {
                "ThingOne": {
                    "foo": int,
                },
            }
        )
    ) as dm:
        msg = dm.ThingOne(1)
        with pytest.raises(AttributeError):
            msg.get_field_message_type("bar")


#############################
## Serialization Edge Cases ##
#############################
def test_bytes_are_json_serializable():
    """Ensure that bytes are json serializable."""
    with temp_data_model(
        make_proto_def(
            {
                "ByteParty": {
                    "bobject": "bytes",
                },
            }
        )
    ) as dm:
        bytestr = b"json may not like me!"
        msg = dm.ByteParty(bytestr)
        # By default, json serialization doesn't handle bytes; make sure we handle it by default.
        json_msg = msg.to_json()
        assert isinstance(json_msg, str)
        # When we reload back from json, we should still have the same bytestring
        reloaded_msg = dm.ByteParty.from_json(json_msg)
        assert isinstance(reloaded_msg.bobject, bytes)
        assert reloaded_msg.bobject == bytestr


def test_primitive_maps_are_serializable():
    """Ensure that we correctly handle primitive values for de/serialization."""
    with temp_data_model(
        make_proto_def(
            {
                "MapParty": {
                    "mobject": "Dict[str, str]",
                },
            }
        )
    ) as dm:
        msg = dm.MapParty({"foo": "bar"})
        # Make sure we can proto and back
        recon_msg = dm.MapParty.from_proto(msg.to_proto())
        assert isinstance(recon_msg, dm.MapParty)
        assert recon_msg.mobject["foo"] == "bar"
        # Make sure we can json and back
        recon_msg = dm.MapParty.from_json(msg.to_json())
        assert isinstance(recon_msg, dm.MapParty)
        assert recon_msg.mobject["foo"] == "bar"


def test_nonprimitive_maps_are_serializable():
    """Ensure that we correctly handle primitive values for de/serialization."""
    with temp_data_model(
        make_proto_def(
            {
                "ComplexType": {
                    "foo": int,
                },
                "MapParty": {
                    "mobject": "Dict[str, ComplexType]",
                },
            }
        )
    ) as dm:
        nested_msg = dm.ComplexType(100)
        msg = dm.MapParty({"foo": nested_msg})
        # Make sure we can proto and back
        recon_msg = dm.MapParty.from_proto(msg.to_proto())
        assert isinstance(recon_msg, dm.MapParty)
        assert recon_msg.mobject["foo"].foo == 100
        # Make sure we can json and back
        recon_msg = dm.MapParty.from_json(msg.to_json())
        assert isinstance(recon_msg, dm.MapParty)
        assert recon_msg.mobject["foo"].foo == 100


def test_union_of_lists_are_serializable():
    """Ensure that we correctly handle union of list values for de/serialization."""
    with temp_data_model(
        make_proto_def(
            {
                "ComplexType": {
                    "foo": "Union[List[str], List[bool]]",
                    "bar": "Union[List[str], List[bool], int]",
                },
            }
        )
    ) as dm:
        # the below is same as: msg = dm.ComplexType(foo=["hello", "world"], bar=1)
        msg = dm.ComplexType(["hello", "world"], 1)
        # Make sure we can proto and back
        recon_msg = dm.ComplexType.from_proto(msg.to_proto())
        assert isinstance(recon_msg, dm.ComplexType)
        assert recon_msg.foo == ["hello", "world"]
        assert recon_msg.bar == 1
        # Make sure we can json and back
        recon_msg = dm.ComplexType.from_json(msg.to_json())
        assert isinstance(recon_msg, dm.ComplexType)
        assert recon_msg.foo == ["hello", "world"]
        assert recon_msg.bar == 1


###############
## From Data ##
###############



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluate quality of models.
"""

# Standard
from dataclasses import dataclass
from typing import Dict, Optional
import enum
import math
import re

# First Party
import alog

# Local
from ..exceptions import error_handler

log = alog.use_channel("TLKIT")
error = error_handler.get(log)


class EvalTypes(enum.Enum):
    """Enum that contains set of all possible evaluation types."""

    SINGLELABEL_MULTICLASS = 1
    MULTILABEL_MULTICLASS = 2
    MULTILABEL_MULTICLASS_HIERARCHICAL = 3


@dataclass
class F1Metrics:
    true_positive: Optional[int] = None
    false_positive: Optional[int] = None
    false_negative: Optional[int] = None
    precision: Optional[float] = None
    recall: Optional[float] = None
    f1: Optional[float] = None


@dataclass
class F1MetricsContainer:
    per_class_confusion_matrix: Dict[str, F1Metrics]
    macro_metrics: F1Metrics
    micro_metrics: F1Metrics


class QualityEvaluator:
    """Class that holds all evaluation logic for now. May eventually be broken up into
    subclasses."""

    def __init__(self, gold, pred):
        self.gold = gold
        self.pred = pred

    def run(
        self,
        evaluation_type,
        find_label_func=None,
        find_label_data_func=None,
        detailed_metrics=False,
        labels=None,
        partial_match_metrics=False,
        max_hierarchy_levels=3,
    ):
        """Main entry point for evaluation.

        Args:
            evaluation_type (str): Which type of evaluation to run. Only a few
                are currently supported.
            find_label_func: function to fetch labels from any one prediction, used in
                multiclass multilabel evaluation.
                eg: if a prediction is of form (token, label), this function should be
                able to tell us how to extract the class labels from the prediction, in
                this case return the second element of the tuple.
            find_label_data_func: function to fetch predictions that belongs to a certain label,
                used only in multiclass multilabel eval type, e.g., if predictions for a data
                example looks like [(tok1, labX), (tok2, labY), (tok3, labX)], then
                the function should be able to return all predictions with a given label - labX
                return should look like [(tok1, labX), (tok3, labX)]
            detailed_metrics: flag to indicate whether or not you want detailed metrics
                              (currently only for multiclass multilabel eval type)
                              Detailed metrics give us metrics for every example, and
                              metrics using a custom partial match function
            labels: list (Optional, defaults to None)
                Optional list of class labels to evaluate quality on. By default evaluation is done
                over all class labels. Using this, you can explicitly mention only a subset of
                labels to include in the quality evaluation.
            partial_match_metrics: flag to indicate whether or not you want partial match
                                   micro avg metrics.
                                   (currently only for multiclass multilabel eval type)
            max_hierarchy_levels (int): Used in hierarchical multilabel
                multiclass evaluation only. The number of levels in the
                hierarchy to run model evaluation on, in addition to complete
                matches.

        Returns:
            dict: Full results from evaluation on dataset and model.
        """
        if evaluation_type == EvalTypes.MULTILABEL_MULTICLASS:
            return self.multilabel_multiclass_evaluation(
                find_label_func,
                find_label_data_func,
                labels,
                detailed_metrics,
                partial_match_metrics,
            )
        if evaluation_type == EvalTypes.SINGLELABEL_MULTICLASS:
            return self.singlelabel_multiclass_evaluation(labels)
        if evaluation_type == EvalTypes.MULTILABEL_MULTICLASS_HIERARCHICAL:
            return self.multilabel_multiclass_hierarchical_evaluation(
                find_label_func, find_label_data_func, max_hierarchy_levels
            )

        error(
            "<COR81451123E>",
            ValueError("Unknown evaluation_type: {0}".format(evaluation_type)),
        )

    def singlelabel_multiclass_evaluation(self, labels=None) -> dict:
        """Obtain results of evaluation for a single-label, multi-class model.

        Args:
            Note: here class should be initialized with gold and pred in the following format
            self.gold (list): list of gold set labels for every example, where each example
                can have only one label eg: ['label1','label2', 'label3','label4']
            self.pred (list): Predicted-by-the-model set labels for every example.
            labels: list (Optional, defaults to None)
                Optional list of class labels to evaluate quality on. By default evaluation is done
                over all class labels. Using this, you can explicitly mention only a subset of
                labels to include in the quality evaluation.

        Returns:
            dict: Dictionary looks like: { 'per_class_confusion_matrix':
                {'entity_type': {'true_positive': int ...}} 'macro_precision': 0
                <= float <= 1, 'macro_recall': 0 <= float <= 1, 'macro_f1': 0 <=
                float <= 1, 'micro_precision': 0 <= float <= 1,, 'micro_recall':
                0 <= float <= 1,, 'micro_f1': 0 <= float <= 1, 'overall_tp':
                int, 'overall_fp': int, 'overall_fn': int


                }
        """

        gold, pred = self.gold, self.pred
        assert len(gold) == len(
            pred
        ), "Length of gold and predicted datasets does not match"

        per_class_confusion_matrix = {}

        for gold_label, pred_label in zip(gold, pred):
            if gold_label not in per_class_confusion_matrix and (
                labels is None or gold_label in labels
            ):
                per_class_confusion_matrix[gold_label] = F1Metrics(
                    true_positive=0,
                    false_positive=0,
                    false_negative=0,
                    precision=0.0,
                    recall=0.0,
                    f1=0.0,
                )
            if pred_label not in per_class_confusion_matrix and (
                labels is None or pred_label in labels
            ):
                per_class_confusion_matrix[pred_label] = F1Metrics(
                    true_positive=0,
                    false_positive=0,
                    false_negative=0,
                    precision=0.0,
                    recall=0.0,
                    f1=0.0,
                )
            # true positive
            if gold_label == pred_label and (labels is None or gold_label in labels):
                per_class_confusion_matrix[gold_label].true_positive += 1
            else:
                if labels is None or pred_label in labels:
                    per_class_confusion_matrix[pred_label].false_positive += 1
                if labels is None or gold_label in labels:
                    per_class_confusion_matrix[gold_label].false_negative += 1

        calc_metrics = QualityEvaluator.calc_metrics_from_confusion_matrix(
            per_class_confusion_matrix
        )
        metrics_out = QualityEvaluator.convert_F1MetricsContainer_to_dict(calc_metrics)
        return metrics_out

    def multilabel_multiclass_evaluation(
        self,
        find_label_func,
        find_label_data_func,
        labels=None,
        detailed_metrics=False,
        partial_match_metrics=False,
        use_labels_for_matching=False,
    ) -> dict:
        """Obtain results of evaluation for a multi-label, multi-class model.

        Args:
            Note: here class should be initialized with gold and pred in the following format
            self.gold (list(list)): list of gold set labels for every example eg:
                [['label1','label2'], ['label1', 'label4']]
            self.pred (list(list)): Predicted-by-the-model set labels for every example.
            find_label_func: function to fetch labels from any one prediction
            find_label_data_func: function to fetch data that belongs to a certain class
            labels: list (Optional, defaults to None)
                Optional list of class labels to evaluate quality on. By default evaluation is done
                over all class labels. Using this, you can explicitly mention only a subset of
                labels to include in the quality evaluation.
            detailed_metrics: flag to indicate whether or not you want detailed metrics
                              Detailed metrics give us metrics for every example, and
                              metrics using a custom partial match function
            partial_match_metrics: flag to indicate whether or not you want partial match
                                   micro avg metrics.
            use_labels_for_matching (bool): Indicates whether or not we should
                use the output of find_label_func for metric computations, or
                the raw data tuples.

        Returns:
            dict: Dictionary looks like: { 'per_class_confusion_matrix':
                {'entity_type': {'true_positive': int ...}} 'macro_precision': 0
                <= float <= 1, 'macro_recall': 0 <= float <= 1, 'macro_f1': 0 <=
                float <= 1, 'micro_precision': micro_precision, 'micro_recall':
                micro_recall, 'micro_f1': micro_f1, 'detailed_metrics' :
                {'exact_match_precision'..,'partial_match_precision'}
                'micro_precision_partial_match': 0 <= float <= 1,
                'micro_recall_partial_match': 0 <= float <= 1,
                'micro_f1_partial_match': 0 <= float <= 1 }
        """
        gold, pred = self.gold, self.pred
        assert len(gold) == len(
            pred
        ), "Length of gold and predicted datasets does not match"
        detailed_output = []
        per_class_confusion_matrix = {}
        all_labels = set()
        num_preds_partial_matched = 0
        num_gold_partial_matched = 0
        total_pred = 0
        total_gold = 0
        overall_tp = 0
        overall_fp = 0
        overall_fn = 0
        micro_precision = 0.0
        micro_recall = 0.0
        micro_f1 = 0.0

        micro_metrics_partial = F1Metrics(precision=0.0, recall=0.0, f1=0.0)
        if labels:
            try:
                gold = [
                    [label for label in gold_ex if find_label_func(label) in labels]
                    for gold_ex in gold
                ]
                pred = [
                    [label for label in pred_ex if find_label_func(label) in labels]
                    for pred_ex in pred
                ]
            except NotImplementedError:
                error(
                    "<COR19114599E>",
                    NotImplementedError(
                        "find_label_func must be implemented to use [labels]"
                    ),
                )

        for gold_ex, pred_ex in zip(gold, pred):
            if detailed_metrics:
                precision, recall, f1 = QualityEvaluator.calc_f1_score(gold_ex, pred_ex)
                (
                    partial_precision,
                    partial_recall,
                    partial_f1,
                ) = QualityEvaluator.calc_f1_score(
                    gold_ex, pred_ex, QualityEvaluator.find_partial_matches
                )
                instances = {
                    "exact_match_precision": precision,
                    "exact_match_recall": recall,
                    "exact_match_f1": f1,
                    "partial_match_precision": partial_precision,
                    "partial_match_recall": partial_recall,
                    "partial_match_f1": partial_f1,
                }
                detailed_output.append(instances)
            try:
                gold_labels = set(map(find_label_func, gold_ex))
                pred_labels = set(map(find_label_func, pred_ex))
                # get per-class information if possible
                all_labels = gold_labels.union(pred_labels)

            except NotImplementedError:
                # If find_label_func raises NotImplementedError, we can't do label-based matching.
                # In this case we need to fall back to set operations on the raw data tuples.
                log.info(
                    "INFO: find_label_func not implemented for this module type - falling back "
                    "to tuple match!!"
                )
                use_labels_for_matching = False

            for label in all_labels:
                # dictionary initizalization
                if label not in per_class_confusion_matrix:
                    per_class_confusion_matrix[label] = F1Metrics(
                        true_positive=0,
                        false_positive=0,
                        false_negative=0,
                        precision=0.0,
                        recall=0.0,
                        f1=0.0,
                    )
                # build confusion matrix
                pred_label_data = set(find_label_data_func(pred_ex, label))
                gold_label_data = set(find_label_data_func(gold_ex, label))

                # true positive
                per_class_confusion_matrix[label].true_positive += len(
                    gold_label_data.intersection(pred_label_data)
                )

                # false positive
                per_class_confusion_matrix[label].false_positive += len(
                    pred_label_data - gold_label_data
                )

                # false negative
                per_class_confusion_matrix[label].false_negative += len(
                    gold_label_data - pred_label_data
                )
            if use_labels_for_matching:
                gold_ex_set = gold_labels
                pred_ex_set = pred_labels
            else:
                # In case the user did not specify how to obtain class labels,
                # we can still calculate micro avg using sum of true positives,
                # false positives etc over all examples (over all classes)
                # We should deprecate this section in next major release
                gold_ex_set = set(gold_ex)
                pred_ex_set = set(pred_ex)
                overall_tp += len(gold_ex_set.intersection(pred_ex_set))
                overall_fp += len(pred_ex_set - gold_ex_set)
                overall_fn += len(gold_ex_set - pred_ex_set)
                # Calculate micro average metrics
                # Micro precision =
                # no. of correct precisions over all classes / no. of total predictions
                if not math.isclose(overall_tp + overall_fp, 0):
                    micro_precision = overall_tp / (overall_tp + overall_fp)
                # Micro recall = no. of correct precisions over all classes / no. of true samples
                if not math.isclose(overall_tp + overall_fn, 0):
                    micro_recall = overall_tp / (overall_tp + overall_fn)
                # Micro avg F1 = harmonic mean of precision and recall
                if not math.isclose(micro_precision + micro_recall, 0):
                    micro_f1 = (2.0 * micro_precision * micro_recall) / (
                        micro_precision + micro_recall
                    )

            if partial_match_metrics:
                gold_matched, preds_matched = QualityEvaluator.find_partial_matches(
                    gold_ex_set, pred_ex_set
                )
                num_preds_partial_matched += len(preds_matched)
                num_gold_partial_matched += len(gold_matched)
                total_pred += len(pred_ex_set)
                total_gold += len(gold_ex_set)

        calc_metrics = QualityEvaluator.calc_metrics_from_confusion_matrix(
            per_class_confusion_matrix
        )

        # This section should be deprecated with future refactors
        if not use_labels_for_matching:
            log.warning(
                "WARNING: Only Micro_avg metrics could be calculated based on the information "
                "available for this module type."
            )
            calc_metrics.micro_metrics.precision = micro_precision
            calc_metrics.micro_metrics.recall = micro_recall
            calc_metrics.micro_metrics.f1 = micro_f1
            calc_metrics.micro_metrics.true_positive = overall_tp
            calc_metrics.micro_metrics.false_positive = overall_fp
            calc_metrics.micro_metrics.false_negative = overall_fn

        metrics_out = QualityEvaluator.convert_F1MetricsContainer_to_dict(calc_metrics)
        # This flag only controls calculation of micro average partial match metrics
        # Detailed metrics flag calculates partial match metrics per data row
        if partial_match_metrics:
            # Calculate micro average partial match metrics
            # Micro precision = Number of matched predictions / Number predicted
            # Micro precision = Fraction of retrieved instances that are relevant
            if total_pred > 0:
                micro_metrics_partial.precision = num_preds_partial_matched / total_pred
            # Micro recall = Number of matched gold / Number in gold set
            # Micro recall = Fraction of relevant instances that are retrieved
            if total_gold > 0:
                micro_metrics_partial.recall = num_gold_partial_matched / total_gold
            # Micro avg F1 = harmonic mean of precision and recall
            if not math.isclose(
                micro_metrics_partial.precision + micro_metrics_partial.recall, 0
            ):
                micro_metrics_partial.f1 = (
                    2.0 * micro_metrics_partial.precision * micro_metrics_partial.recall
                ) / (micro_metrics_partial.precision + micro_metrics_partial.recall)

        metrics_out["detailed_metrics"] = detailed_output
        metrics_out["micro_precision_partial_match"] = micro_metrics_partial.precision
        metrics_out["micro_recall_partial_match"] = micro_metrics_partial.recall
        metrics_out["micro_f1_partial_match"] = micro_metrics_partial.f1

        return metrics_out

    def multilabel_multiclass_hierarchical_evaluation(
        self,
        find_label_func_builder,
        find_label_data_func_builder,
        max_hierarchy_levels=3,
    ) -> dict:
        """Evaluate multilabel/multiclass over a hierarchy, e.g., for ESA categories. This method
        Evaluates over a set number of hierarchy levels.

        Because each level in the hierarchy needs to be able to compare and extract differently,
        we use builder funcs that create the appropriate functions for a given level of the
        hierarchy.

        Args:
            find_label_func_builder (function): A function that takes in a level
                number (or None if full hierarchy) and returns a find_label_func
                for this level that can be passed to the multilabel multiclass
                evaluator.
            find_label_data_func_builder (function): A function that takes in a
                level number (or None if full hierarchy) and returns a
                find_label_data_func for this level that can be passed to the
                multilabel multiclass evaluator.
            max_hierarchy_levels (int): The number of levels to run in the
                hierarchy, in addition to complete match.
        Returns:
            dict: Dictionary, where each key is a level number, or 'FULL', and
                maps to the dict returned by multilabel_multiclass_evaluation
                for that level of the hierarchy.
        """
        metrics = {}
        # Levels are None [FULL], and 1...n, where n is the deepest level in the hierarchy (for now,
        # needs to be manually determined by the user).
        # pylint: disable=unnecessary-comprehension
        levels = [None] + [level for level in range(1, max_hierarchy_levels + 1)]
        for level in levels:
            # Get the find_label_func/find_label_data_func for this level in the hierarchy
            find_label_func = find_label_func_builder(level)
            find_label_data_func = find_label_data_func_builder(level)
            # Get the appropriate dictionary key - Fall is a bit more descriptive than None
            dict_key = "level_{}".format(level) if level is not None else "level_all"
            # NOTE: We use label matching for computing our metrics here. This means that we
            # compare the outputs of find_label_func on gold/pred examples to get our metrics
            # instead of the gold/pred example tuples themselves. The reason that we generally
            # want to do this is that the examples have the full labels, but we need to slice out
            # just part of the hierarchical label to consider each level.
            metrics[dict_key] = self.multilabel_multiclass_evaluation(
                find_label_func, find_label_data_func, use_labels_for_matching=True
            )
        return metrics

    @staticmethod
    def calc_f1_score(gold, pred, match_fun=None):
        """Calculates F1 score
        Args:
            gold (list): List of gold annotations
            pred (list): List of predictions
            match_fun: Function that finds the matches and returns tuple of matched gold, preds
        Returns:
            tuple: Precision, Recall, F1 score
        """
        if match_fun:
            # In case of partial match, matched predictions need not equal matched gold
            # Two predictions can match one gold, while another gold may not be matched
            try:
                matched_gold, matched_preds = match_fun(gold, pred)
            except (ValueError, TypeError):
                error(
                    "<COR19474599E>",
                    ValueError("Match function not returning expected tuple format"),
                )
        else:
            matched_preds = set(gold).intersection(set(pred))
            matched_gold = matched_preds

        num_correct_preds, num_pred, num_gold, num_correct_gold = (
            len(matched_preds),
            len(pred),
            len(gold),
            len(matched_gold),
        )

        # precision == Fraction of relevant instances among retrieved instances
        # If we could match 3 predictions with gold out of 4 predictions, precision = 3/4
        precision = num_correct_preds / num_pred if num_pred > 0 else 0.0
        # recall == Fraction of retrieved instances among relevant instances
        # If we could match/retrieve only one gold out of 3, recall = 1/3
        recall = num_correct_gold / num_gold if num_gold > 0 else 0.0
        # f1 == harmonic_mean(precision, recall)
        f1 = (
            (2 * precision * recall) / (precision + recall)
            if precision != 0 and recall != 0
            else 0.0
        )
        return precision, recall, f1

    @staticmethod
    def find_partial_matches(groundtruth, prediction):
        """Function to do find partial match between predicted phrases and the ground truth.
           partial match means a complete predicted phrase is a part of any ground truth phrase or
           a complete ground truth phrase is a part of any predicted phrase.
           Overlaps are not considered.

        Args:
            groundtruth (list): Groundtruth data
            prediction (list): Predictions returned by the model

        Returns:
            tuple: gold_matched: set, pred_matched: set gold annotations that
                were matched Predictions that partially or fully matched with
                groundtruth
        """

        gold_matched = set()
        pred_matched = set()

        for ground_truth_phrase in groundtruth:
            for predicted_phrase in prediction:
                pd_compiler = re.compile(r"\b%s\b" % re.escape(predicted_phrase), re.I)
                # Checks if prediction is part of groundtruth
                if pd_compiler.search(ground_truth_phrase):
                    pred_matched.add(predicted_phrase)
                    gold_matched.add(ground_truth_phrase)
                else:
                    # Checks if groundtruth is part of prediction
                    gt_compiler = re.compile(
                        r"\b%s\b" % re.escape(ground_truth_phrase), re.I
                    )
                    if gt_compiler.search(predicted_phrase):
                        pred_matched.add(predicted_phrase)
                        gold_matched.add(ground_truth_phrase)

        return gold_matched, pred_matched

    @staticmethod
    def calc_metrics_from_confusion_matrix(
        per_class_confusion_matrix: Dict[str, F1Metrics]
    ) -> F1MetricsContainer:
        """Function to calculate precision, recall, F1 metrics using a confusion matrix containing
           statistics per class label.

        Args:
            per_class_confusion_matrix (Dict[str, F1Metrics]): Dictionary of
                 statistics per class label. Class labels are keys for the
                 dictionary. For each class label, there should be a F1Metrics
                 class object with values true positive, false_positive ,
                 false_negative representating the count of these per class. The
                 dictionary looks like: per_class_confusion_matrix[label] =
                 F1Metrics(true_positive = val 1, false_positive = val 2,
                 false_negative = val 3)

        Returns:
            Returns:
            metrics_summary: F1MetricsContainer
            An instance of F1MetricsContainer dataclass containing summary of F1 metrics
        """
        macro_metrics = F1Metrics(precision=0.0, recall=0.0, f1=0.0)
        micro_metrics = F1Metrics(
            true_positive=0,
            false_positive=0,
            false_negative=0,
            precision=0.0,
            recall=0.0,
            f1=0.0,
        )
        num_classes = len(per_class_confusion_matrix)
        # Compute metrics per label
        for label in per_class_confusion_matrix:
            tp = per_class_confusion_matrix[label].true_positive
            fp = per_class_confusion_matrix[label].false_positive
            fn = per_class_confusion_matrix[label].false_negative
            # Calculate precision of a label X = \
            # no. of correct predictions of X / no. of predictions of X
            if not math.isclose(tp + fp, 0):
                per_class_confusion_matrix[label].precision = tp / (tp + fp)
            # Calculate recall of label X = \
            # no. of correct predictions of X / no. of true samples of X
            if not math.isclose(tp + fn, 0):
                per_class_confusion_matrix[label].recall = tp / (tp + fn)
            prec = per_class_confusion_matrix[label].precision
            recall = per_class_confusion_matrix[label].recall
            # Calculate F1 score of label X = harmonic mean of precision and recall of X
            if not math.isclose(prec + recall, 0):
                per_class_confusion_matrix[label].f1 = (2.0 * prec * recall) / (
                    prec + recall
                )

            micro_metrics.true_positive += tp
            micro_metrics.false_positive += fp
            micro_metrics.false_negative += fn

            macro_metrics.precision += per_class_confusion_matrix[label].precision
            macro_metrics.recall += per_class_confusion_matrix[label].recall
            macro_metrics.f1 += per_class_confusion_matrix[label].f1

        # Macro average metrics = average of metrics over all classes
        if num_classes > 0:
            macro_metrics.precision = macro_metrics.precision / num_classes
            macro_metrics.recall = macro_metrics.recall / num_classes
            macro_metrics.f1 = macro_metrics.f1 / num_classes

        # Calculate micro average metrics
        # Micro precision = no. of correct precisions over all classes / no. of total predictions
        if not math.isclose(
            micro_metrics.true_positive + micro_metrics.false_positive, 0
        ):
            micro_metrics.precision = micro_metrics.true_positive / (
                micro_metrics.true_positive + micro_metrics.false_positive
            )
        # Micro recall = no. of correct precisions over all classes / no. of true samples
        if not math.isclose(
            micro_metrics.true_positive + micro_metrics.false_negative, 0
        ):
            micro_metrics.recall = micro_metrics.true_positive / (
                micro_metrics.true_positive + micro_metrics.false_negative
            )
        # Micro avg F1 = harmonic mean of precision and recall
        if not math.isclose(micro_metrics.precision + micro_metrics.recall, 0):
            micro_metrics.f1 = (
                2.0 * micro_metrics.precision * micro_metrics.recall
            ) / (micro_metrics.precision + micro_metrics.recall)

        metrics_summary = F1MetricsContainer(
            per_class_confusion_matrix, macro_metrics, micro_metrics
        )
        return metrics_summary

    # pylint: disable=no-self-argument
    def convert_F1MetricsContainer_to_dict(metrics_summary: F1MetricsContainer) -> dict:
        """
        Args:
            metrics_summary (F1MetricsContainer): An object of dataclass
                 F1MetricsContainer

        Returns:
            Returns:
            dict
                Dictionary looks like: {
                    'per_class_confusion_matrix': {'entity_type': {'true_positive': int ...}}
                    'macro_precision': 0 <= float <= 1,
                    'macro_recall': 0 <= float <= 1,
                    'macro_f1': 0 <= float <= 1,
                    'micro_precision': 0 <= float <= 1,,
                    'micro_recall': 0 <= float <= 1,,
                    'micro_f1': 0 <= float <= 1,
                    'overall_tp': int,
                    'overall_fp': int,
                    'overall_fn': int
                }
        """

        for label, obj in metrics_summary.per_class_confusion_matrix.items():
            # Converts the object to dictionary
            metrics_summary.per_class_confusion_matrix[label] = vars(obj)

        out = {"per_class_confusion_matrix": metrics_summary.per_class_confusion_matrix}

        for k, v in vars(metrics_summary.macro_metrics).items():
            out[f"macro_{k}"] = v

        out["micro_precision"] = metrics_summary.micro_metrics.precision
        out["micro_recall"] = metrics_summary.micro_metrics.recall
        out["micro_f1"] = metrics_summary.micro_metrics.f1
        out["overall_tp"] = metrics_summary.micro_metrics.true_positive
        out["overall_fp"] = metrics_summary.micro_metrics.false_positive
        out["overall_fn"] = metrics_summary.micro_metrics.false_negative

        return out



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import json
import os
import pickle

# Third Party
import pytest

# Local
from caikit.core.augmentors import AugmentorBase
from caikit.core.data_model import DataStream
from sample_lib.data_model.sample import SampleInputType, SampleOutputType
import caikit.core

###############################
# Helper functions and fixtures
###############################


def validate_data_stream(data_stream, length, data_item_type, data_item_length=None):
    # Verify lengths and types
    assert isinstance(data_stream, DataStream)
    assert len(data_stream) == length
    assert sum(1 for _ in data_stream) == length

    for data_item in data_stream:
        assert isinstance(data_item, data_item_type)
        if data_item_length is not None:
            assert len(data_item) == data_item_length


def build_test_augmentor(produces_none):
    class TestStreamAugmentor(AugmentorBase):
        """A 'normal' augmentor, which when leveraged correctly, always produces an output per input."""

        augmentor_type = int
        # Indicates that this class should not be picked up by augmentor discovery utilities.
        is_test_augmentor = True

        def __init__(self):
            super().__init__(random_seed=1001, produces_none=produces_none)

        def _augment(self, obj):
            """If produces_none is True, filters out falsy values, i.e., 0s, otherwise adds
            five to the original input.

            Also return None if we pass 13 and set produces_none=False, as a sad test case.
            This should raise a TypeError within the augmentor base class.
            """
            if produces_none and not obj:
                return None
            if not produces_none and obj == 13:
                return None
            return obj + 5

    return TestStreamAugmentor()


## Local fixtures ###
@pytest.fixture
def list_data_stream():
    yield DataStream.from_iterable(["hello", "world", "!"])


###############################
# Tests on various data formats
###############################


def test_list_data_stream(list_data_stream):
    validate_data_stream(list_data_stream, 3, str)


def test_txt_data_stream(sample_text_file):
    validate_data_stream(DataStream.from_txt(sample_text_file), 7, str)


def test_txt_collection_data_stream(sample_text_collection):
    validate_data_stream(DataStream.from_txt_collection(sample_text_collection), 3, str)


def test_json_collection_data_stream(sample_json_collection):
    validate_data_stream(
        DataStream.from_json_collection(sample_json_collection), 3, dict
    )


def test_jsonl_collection_data_stream(sample_jsonl_collection):
    validate_data_stream(
        DataStream.from_jsonl_collection(sample_jsonl_collection), 7, dict
    )


def test_csv_collection_data_stream(sample_csv_collection):
    csv_collection_data_stream = DataStream.from_csv_collection(sample_csv_collection)
    validate_data_stream(csv_collection_data_stream, 6, dict, 2)
    for data_item in csv_collection_data_stream:
        for key, element in data_item.items():
            assert isinstance(key, str)
            assert isinstance(element, str)


def test_csv_data_stream(sample_csv_file_no_headers):
    csv_data_stream = DataStream.from_csv(sample_csv_file_no_headers)
    validate_data_stream(csv_data_stream, 3, list, 2)
    for data_item in csv_data_stream:
        for element in data_item:
            assert isinstance(element, str)


def test_csv_header_data_stream(sample_csv_file):
    csv_header_data_stream = DataStream.from_header_csv(sample_csv_file)
    validate_data_stream(csv_header_data_stream, 3, dict, 2)
    for data_item in csv_header_data_stream:
        for element in data_item:
            assert isinstance(element, str)


def test_json_array_data_stream(sample_json_file):
    json_data_stream = DataStream.from_json_array(sample_json_file)
    with open(sample_json_file) as f:
        assert list(json_data_stream) == json.load(f)


def test_jsonl_array_data_stream(sample_jsonl_file):
    jsonl_data_stream = DataStream.from_jsonl(sample_jsonl_file)
    validate_data_stream(jsonl_data_stream, 4, dict, 2)
    for data_item in jsonl_data_stream:
        for element in data_item:
            assert isinstance(element, str)


def test_jsonl_control_chars_array_data_stream(jsonl_with_control_chars):
    data_stream = DataStream.from_jsonl(jsonl_with_control_chars)
    validate_data_stream(data_stream, 2, dict, 2)
    for data_item in data_stream:
        for element in data_item:
            assert isinstance(element, str)


def test_data_stream_from_jsonl_is_pickleable(tmp_path):
    tmpdir = str(tmp_path)

    data = [1, 2, 3, 4, 5, 6]
    filepath = os.path.join(tmpdir, "foo.jsonl")
    with open(filepath, "w") as f:
        json.dump(data, f)

    stream = DataStream.from_jsonl(filepath)

    pre_pickle_vals = list(stream)
    pickled_stream = pickle.loads(pickle.dumps(stream))
    post_pickle_vals = list(pickled_stream)

    assert pre_pickle_vals == post_pickle_vals


def test_bad_json_stream(tmp_path):
    file_path = os.path.join(str(tmp_path), "bad.json")
    with open(file_path, "w") as fp:
        fp.write(
            """
            text
            Some
            { "a" : 1}
            "a": 1
        """
        )
    with pytest.raises(ValueError, match="Invalid JSON object"):
        DataStream.from_json_array(file_path).peek()


def test_non_json_array_stream(tmp_path):
    file_path = os.path.join(str(tmp_path), "non_array.json")
    with open(file_path, "w") as fp:
        fp.write(
            """
        {
            "number": 1,
            "label": "foo"
        }
        """
        )
    with pytest.raises(ValueError, match="Non-array JSON object"):
        DataStream.from_json_array(file_path).peek()


def test_bad_multipart_file(tmp_path):
    multipart_file = os.path.join(str(tmp_path), "bad_multipart")
    with open(multipart_file, "w") as fp:
        fp.write("content")
        fp.write("\n")
        fp.write("--")
        fp.write("arbitrary")
    with pytest.raises(ValueError, match="file is not multipart"):
        DataStream.from_multipart_file(multipart_file)


def test_invalid_multipart_boundary(tmp_path):
    file_path = os.path.join(str(tmp_path), "bad.multipart")
    with open(file_path, "w") as fp:
        fp.write(
            """
Content-Type: multipart/form-data; boundary=foo

--foo
Content-Disposition: form-data; name=""; filename="my_training_data.json"
Content-Type: application/json

[
    {
        "number": 1,
        "label": "small"
    },
    {
        "number": 1000,
        "label": "large"
    }
]

--bar--
            """
        )
    with pytest.raises(ValueError, match="Invalid form-data"):
        DataStream.from_multipart_file(file_path).peek()


def test_from_file_can_handle_a_json_file(sample_json_file):
    json_stream = DataStream.from_file(sample_json_file)
    validate_data_stream(json_stream, 2, dict)


def test_from_multipart_file_json(sample_multipart_json):
    multipart_stream = DataStream.from_multipart_file(sample_multipart_json)
    validate_data_stream(multipart_stream, 2, dict)


def test_from_multipart_file_csv(sample_multipart_csv):
    multipart_stream = DataStream.from_multipart_file(sample_multipart_csv)
    validate_data_stream(multipart_stream, 2, dict)


def test_from_multipart_file_unsupported_content_type(tmp_path):
    multipart_file = os.path.join(str(tmp_path), "multipart")

    # Still need to create valid multipart file
    with open(multipart_file, "w") as fp:
        fp.write("--")
        fp.write("arbitrary")
        fp.writelines(
            [
                "\n",
                f'Content-Disposition: form-data; name=""; filename="file"\n',
                "Content-Type: text/plain\n",
                "\n",
            ]
        )
        fp.write("content")
        fp.write("\n")
        fp.write("--")
        fp.write("arbitrary")
        fp.write("--")
    with pytest.raises(ValueError, match="Unsupported content type: text/plain"):
        DataStream.from_multipart_file(multipart_file)


###########################
# General data stream tests
###########################


def test_iteration_is_lazy():
    """Verify that iteration over a simple data stream is lazy, i.e., elements are generated
    as they are requested.
    """

    def generator_func():
        generator_func.it = 0
        for i in range(10):
            yield i
            generator_func.it += 1

    counter_stream = DataStream(generator_func)

    for count, _ in enumerate(counter_stream):
        assert generator_func.it == count


def test_generator_not_callable():
    """Verify that a `generator_func` that is not callable raises a `TypeError`."""
    with pytest.raises(TypeError):
        DataStream(42)


def test_empty_generator_func():
    """Verify that an empty data stream has zero columns and zero data items."""

    def generator_func():
        return iter(range(0))

    empty_data_stream = DataStream(generator_func)

    assert len(empty_data_stream) == 0


def test_zip(sample_csv_file_no_headers):
    """Verify that `.zip` correctly combines data stream."""
    data_stream1 = DataStream(lambda: iter(range(3)))
    validate_data_stream(data_stream1, 3, int)

    data_stream2 = DataStream(lambda: iter(range(3, 0, -1)))
    validate_data_stream(data_stream2, 3, int)

    zipped_streams = data_stream1.zip(data_stream2)
    validate_data_stream(zipped_streams, 3, tuple, 2)

    zipped_streams = DataStream.zip(
        data_stream1, data_stream2, DataStream.from_csv(sample_csv_file_no_headers)
    )
    validate_data_stream(zipped_streams, 3, tuple, 3)

    for data_item in zipped_streams:
        assert isinstance(data_item[0], int)
        assert isinstance(data_item[1], int)
        assert isinstance(data_item[2], list)
        assert len(data_item[2]) == 2


def test_zip_different_lengths(sample_text_file, list_data_stream):
    with pytest.raises(ValueError):
        list(DataStream.zip(list_data_stream, DataStream.from_txt(sample_text_file)))


def test_zip_is_lazy():
    """Verify that a simple data stream is still lazily evaluated after `.zip`."""

    def generator_func1():
        generator_func1.it = 0
        for i in range(10):
            yield i
            generator_func1.it += 1

    data_stream1 = DataStream(generator_func1)

    def generator_func2():
        generator_func2.it = 0
        for i in range(10, 20):
            yield i
            generator_func2.it += 1

    data_stream2 = DataStream(generator_func2)

    for count, _ in enumerate(data_stream1.zip(data_stream2)):
        assert generator_func1.it == count
        assert generator_func2.it == count


def test_zip_unequal_lengths():
    """Verify that attempting to zip streams with unequal length results in a value error
    when iterating over the stream.
    """
    data_stream1 = DataStream.from_iterable(range(5))
    data_stream2 = DataStream.from_iterable(range(10))
    zipped_stream = data_stream1.zip(data_stream2)

    with pytest.raises(ValueError):
        list(zipped_stream)


def test_slice():
    """Verify that `.__getitem__` slices data items."""

    def generator_func():
        for i in range(10):
            yield i, i + 1, i + 2, i + 3

    orig_data_stream = DataStream(generator_func)
    validate_data_stream(orig_data_stream, 10, tuple, 4)

    data_stream = orig_data_stream[0]
    validate_data_stream(data_stream, 10, int)
    for count, data_item in enumerate(data_stream):
        assert data_item == count

    data_stream = orig_data_stream[:2]
    validate_data_stream(data_stream, 10, tuple, 2)
    for count, data_item in enumerate(data_stream):
        assert data_item[0] == count
        assert data_item[1] == count + 1

    data_stream = orig_data_stream[1:3]
    validate_data_stream(data_stream, 10, tuple, 2)
    for count, data_item in enumerate(data_stream):
        assert data_item[0] == count + 1
        assert data_item[1] == count + 2


def test_slice_is_lazy():
    """Verify that a simple generator is still lazily evaluated after slicing."""

    def generator_func():
        generator_func.it = 0
        for i in range(10):
            yield i, i + 1, i + 2, i + 3, i + 4
            generator_func.it += 1

    data_stream = DataStream(generator_func)

    for count, _ in enumerate(data_stream[0:2]):
        assert generator_func.it == count


def test_train_test_split():
    """Verify that a source stream can be split into two substreams with the appropriate split
    ratio and no overlapping elements
    """
    # Split the stream using a predefined seed
    src_stream = DataStream.from_iterable(range(0, 1000))
    train_stream, test_stream = src_stream.train_test_split(test_split=0.25, seed=42)
    assert len(train_stream) + len(test_stream) == len(src_stream)
    validate_data_stream(train_stream, 762, int)  # Checks for deterministic split
    validate_data_stream(test_stream, 238, int)  # as well as correct ratio
    # Tests for no intersection between train and test stream
    assert len(set(train_stream).intersection(test_stream)) == 0
    # Tests that there no duplicates in train and test streams
    assert len(set(train_stream)) == len(train_stream)
    assert len(set(test_stream)) == len(test_stream)

    # Verify that split is repeatable given same parameters
    train_stream2, test_stream2 = src_stream.train_test_split(test_split=0.25, seed=42)
    assert len(train_stream2) + len(test_stream2) == len(src_stream)
    validate_data_stream(train_stream2, 762, int)  # Checks for deterministic split
    validate_data_stream(test_stream2, 238, int)  # as well as correct ratio
    assert all([a == b for (a, b) in zip(train_stream, train_stream2)])
    assert all([a == b for (a, b) in zip(test_stream, test_stream2)])

    # Verify that split is different given different seed
    train_stream3, test_stream3 = src_stream.train_test_split(test_split=0.25, seed=501)
    assert len(train_stream3) + len(test_stream3) == len(src_stream)
    validate_data_stream(train_stream3, 754, int)  # Checks for deterministic split
    validate_data_stream(test_stream3, 246, int)  # as well as correct ratio

    # Verify that split is different given different test_split
    train_stream4, test_stream4 = src_stream.train_test_split(test_split=0.75, seed=42)
    assert len(train_stream4) + len(test_stream4) == len(src_stream)
    validate_data_stream(train_stream4, 257, int)  # Checks for deterministic split
    validate_data_stream(test_stream4, 743, int)  # as well as correct ratio


def test_shuffle():
    """Verify that a source stream can be shuffled"""

    # Shuffle the stream using a predefined seed
    src_stream = DataStream.from_iterable(range(0, 1000))
    shuffled_stream = src_stream.shuffle(buffer_size=200, seed=42)
    validate_data_stream(shuffled_stream, len(src_stream), int)
    assert set(shuffled_stream) == set(src_stream)

    # Verify that shuffle is repeatable given same parameters
    shuffled_stream2 = src_stream.shuffle(buffer_size=200, seed=42)
    validate_data_stream(shuffled_stream2, len(src_stream), int)
    assert list(shuffled_stream) == list(shuffled_stream2)

    # Verify that shuffle is different given different seed
    shuffled_stream3 = src_stream.shuffle(buffer_size=200, seed=452)
    validate_data_stream(shuffled_stream3, len(src_stream), int)
    assert list(shuffled_stream) != list(shuffled_stream3)

    # Verify that shuffle is different given different buffer_size
    shuffled_stream4 = src_stream.shuffle(buffer_size=500, seed=42)
    validate_data_stream(shuffled_stream4, len(src_stream), int)
    assert list(shuffled_stream) != list(shuffled_stream4)

    # Verify that shuffle is different given different buffer_size
    src_stream = DataStream.from_iterable(range(0, 100))
    shuffled_stream = src_stream.shuffle(buffer_size=500, seed=42)
    validate_data_stream(shuffled_stream, len(src_stream), int)
    assert set(shuffled_stream) == set(src_stream)
    assert list(shuffled_stream) != list(src_stream)


def test_chain(
    sample_csv_file_no_headers,
    sample_text_file,
    sample_text_collection,
    list_data_stream,
):
    csv_data_stream = DataStream.from_csv(sample_csv_file_no_headers)
    """Verify that data streams can be chained together."""
    cat_csv_pipline = DataStream.chain(
        csv_data_stream, csv_data_stream, csv_data_stream
    )
    validate_data_stream(cat_csv_pipline, 9, list, 2)

    cat_txt_data_stream = DataStream.chain(
        list_data_stream,
        DataStream.from_txt(sample_text_file),
        DataStream.from_txt_collection(sample_text_collection),
    )
    validate_data_stream(cat_txt_data_stream, 13, str)


def test_chain_is_lazy():
    """Verify that chained data streams are lazily evaluated."""

    def generator_func1():
        for i in range(10):
            generator_func1.it = i
            yield i, i + 1

    data_stream1 = DataStream(generator_func1)

    def generator_func2():
        for i in range(10, 20):
            generator_func1.it = i
            yield i, i + 1

    data_stream2 = DataStream(generator_func2)

    chained_data_stream = data_stream1.chain(data_stream2)
    validate_data_stream(chained_data_stream, 20, tuple, 2)

    for count, _ in enumerate(chained_data_stream):
        assert generator_func1.it == count


def test_add(sample_text_collection, list_data_stream):
    """Verify that the + operator chains data streams."""
    add_data_stream = (
        list_data_stream
        + list_data_stream
        + DataStream.from_txt_collection(sample_text_collection)
    )
    validate_data_stream(add_data_stream, 9, str)


def test_eager_is_not_lazy():
    """After running `.eager`, the stream is placed into memory and no longer lazy evaluated."""

    def generator_func():
        for i in range(10):
            generator_func.it = i
            yield i, i + 1

    data_stream = DataStream(generator_func)

    for _ in data_stream.eager():
        assert generator_func.it == 9


def test_map():
    """Verify that a simple application of the map functino works as expected."""
    # stream over integer range
    int_stream = DataStream(lambda: iter(range(10)))

    # map function returns a tuple with both the original value and also the value plus one
    int_plus_one_stream = int_stream.map(lambda data_item: (data_item, data_item + 1))

    for count, data_item in enumerate(int_plus_one_stream):
        assert data_item[0] == count
        assert data_item[1] == count + 1


def test_map_is_lazy():
    """Verify that lazy evaluation occurs after a `.map`."""

    def generator_func():
        generator_func.it = 0
        for i in range(10):
            yield i
            generator_func.it += 1

    int_stream = DataStream(generator_func)

    for count, data_item in enumerate(int_stream):
        assert data_item == count
        assert generator_func.it == count


def test_filter_empty_lines(sample_text_file):
    """Verify that a simple filter can remove empty lines of text."""
    no_empty_txt_stream = DataStream.from_txt(sample_text_file).filter()
    assert len(no_empty_txt_stream) == 4


def test_filter_is_lazy():
    """Verify that lazy evaluation occurs after a `.filter`."""

    def generator_func():
        generator_func.it = 0
        for i in range(10):
            yield i
            generator_func.it += 1

    int_stream = DataStream(generator_func)
    odd_int_stream = int_stream.filter(lambda data_item: data_item % 2)
    validate_data_stream(odd_int_stream, 5, int)

    for count, data_item in enumerate(odd_int_stream):
        assert data_item == count * 2 + 1


def test_dummy_stream(good_model_path, sample_csv_file):
    """Verify that we can use streams to generate dummy predictions."""
    dummy_model = caikit.core.load(good_model_path)

    stream_len = 7
    sample_input_stream = DataStream.from_iterable(
        [SampleInputType(name="gabe")] * stream_len
    )
    dummy_stream = dummy_model.stream(sample_input_stream)
    validate_data_stream(dummy_stream, stream_len, SampleOutputType)


def test_only_one_stream_allowed(
    sample_csv_file_no_headers, good_model_path
):  # , mock_run):
    """Verify that we can't call module.stream() with multiple DataStream args."""
    stream = DataStream.from_iterable([1, 2, 3, 4, 5])

    dummy_model = caikit.core.load(good_model_path)
    with pytest.raises(ValueError):
        # NOTE: Even though this is too many args for .run() to handle, it's okay; this
        # will fail before it gets to .run(). In the event that .run() is called with the
        # wrong number of positional arguments, it'll throw a TypeError, which fails this test.
        dummy_model.stream(stream, stream)


def test_pipe_stream(good_model_path):
    """Verify that we can use streams to generate dummy predictions using `|` syntax."""
    dummy_model = caikit.core.load(good_model_path)
    stream_len = 7
    sample_input_stream = DataStream.from_iterable(
        [SampleInputType(name="gabe")] * stream_len
    )

    dummy_stream = sample_input_stream | dummy_model
    validate_data_stream(dummy_stream, stream_len, SampleOutputType)


##################
# Augmentor tests
##################


def test_augment_enforces_augmentor_type_checking():
    """Ensure that we need to pass an augmentor to augment a stream."""
    input_stream = DataStream.from_iterable([1, 0, 3, 2])
    with pytest.raises(TypeError):
        input_stream.augment("garbage", 1)


def test_augment_validates_aug_cycles_properly():
    """Ensure that we properly validate aug_cycles when augmenting."""
    input_stream = DataStream.from_iterable([1, 0, 3, 2])
    aug = build_test_augmentor(True)
    with pytest.raises(TypeError):
        input_stream.augment(aug, "Garbage")
    with pytest.raises(ValueError):
        input_stream.augment(aug, 0)


def test_augment_valides_enforce_determinism_properly():
    """Make sure that we throw if enforce_determinism is not a bool."""
    input_stream = DataStream.from_iterable([1, 0, 3, 2])
    aug = build_test_augmentor(True)
    with pytest.raises(TypeError):
        input_stream.augment(aug, 2, enforce_determinism=4)


def test_augment_validates_post_func_callback():
    """Ensure that the post augmentation callback needs to be a function."""
    input_stream = DataStream.from_iterable([1, 0, 3, 2])
    aug = build_test_augmentor(True)
    with pytest.raises(TypeError):
        input_stream.augment(aug, 2, post_augment_func="not a function")


def test_augmentor_that_returns_none_when_expected():
    """Test that if an augmentor is allowed to return None, none values drop."""
    input_sequence = [1, 0, 3, 2]
    output_sequence = [1, 0, 3, 2, 6, 8, 7]
    aug = build_test_augmentor(True)
    actual_sequence = list(DataStream.from_iterable(input_sequence).augment(aug, 1))
    assert output_sequence == actual_sequence


def test_augmentor_that_returns_none_when_not_expected():
    """Test that we raise if an augmentor produces None when it is not supposed to."""
    input_sequence = [13]
    aug = build_test_augmentor(False)
    # Explode because we said this augmentor should never raise None, and we pass unlucky 13 :(
    with pytest.raises(TypeError):
        list(DataStream.from_iterable(input_sequence).augment(aug, 1))


def test_index_targeting():
    """Test that index targeting works as expected."""
    input_sequence = [(1, "foo"), (0, "bar")]
    output_sequence = [(1, "foo"), (0, "bar"), (6, "foo"), (5, "bar")]
    input_stream = DataStream.from_iterable(input_sequence)
    aug = build_test_augmentor(False)
    actual_sequence = list(input_stream.augment(aug, 1, augment_index=0))
    assert output_sequence == actual_sequence


def test_out_of_bounds_index_target():
    """Test that if we specify an out of bounds index target, we explode correctly."""
    input_sequence = [(1, "foo"), (0, "bar")]
    input_stream = DataStream.from_iterable(input_sequence)
    aug = build_test_augmentor(False)
    with pytest.raises(IndexError):
        list(input_stream.augment(aug, 1, augment_index=2))


def test_post_augment_func():
    """Test that we can apply a post augmentation callback."""
    callback = lambda val: str(val)
    input_sequence = [1, 0, 3, 2]
    output_sequence = ["1", "0", "3", "2", "6", "5", "8", "7"]
    input_stream = DataStream.from_iterable(input_sequence)
    aug = build_test_augmentor(False)
    actual_sequence = list(input_stream.augment(aug, 1, post_augment_func=callback))
    assert output_sequence == actual_sequence


def test_post_augment_func_with_index_targeting():
    """Test that we can apply a post augmentation callback with targeted indices."""
    input_sequence = [(1, "foo"), (0, "bar")]
    # Callback will stringify the values that we get back
    callback = lambda val: str(val)
    output_sequence = [("1", "foo"), ("0", "bar"), ("6", "foo"), ("5", "bar")]
    input_stream = DataStream.from_iterable(input_sequence)
    aug = build_test_augmentor(False)
    actual_sequence = list(
        input_stream.augment(aug, 1, augment_index=0, post_augment_func=callback)
    )
    assert output_sequence == actual_sequence



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Local loader implementation for loading modules from disk.
Contains recursive functions for loading modules saved inside modules.
"""

# Standard
from typing import Union
import os

# First Party
import alog

# Local
from ..exceptions import error_handler
from .config import ModuleConfig
from caikit import core

log = alog.use_channel("MODULE_LOAD")
error = error_handler.get(log)


class ModuleLoader:
    MODULE_PATHS_KEY = "module_paths"

    def __init__(self, model_path: Union[str, ModuleConfig]):
        """Construct a new module loader.

        Args:
            model_path (Union[str, ModuleConfig]): The path to the directory
                where the model is to be loaded from, or a preloaded config.
        """
        self.config = ModuleConfig.load(model_path)
        self.model_path = self.config.model_path

    def load_arg(self, arg):
        """Extract arg value from the loaded model's config"""
        return getattr(self.config, arg)

    def load_args(self, *args):
        """Extract values from the loaded model's config"""
        return tuple(getattr(self.config, arg) for arg in args)

    def load_module(self, module_paths_key, load_singleton=False):
        """Load a CaikitCore module from a module config.module_paths specification.

        Args:
            module_paths_key (str): key in `config.module_paths` looked at to
                load a module
            load_singleton (bool): singleton load flag to pass to individual
                module loads
        """
        # Load module from a given relative path
        if "module_paths" not in self.config:
            error(
                "<COR08580509E>", KeyError("Missing `module_paths` in module config!")
            )

        if module_paths_key not in self.config.module_paths:
            error(
                "<COR22069088E>",
                KeyError(
                    "Missing required {} key in config.module_paths!".format(
                        module_paths_key
                    )
                ),
            )

        module_path = os.path.join(
            self.model_path, self.config.module_paths[module_paths_key]
        )
        return core.load(module_path, load_singleton=load_singleton)

    def load_module_list(self, module_paths_key):
        """Load a list of CaikitCore module from a workflow config.module_paths specification.

        Args:
            module_paths_key (str): key in `config.module_paths` looked at to
                load a list of modules

        Returns:
            list: list of loaded modules
        """
        # Load module from a given relative path
        # Can be updated to load from a module key
        if self.MODULE_PATHS_KEY not in self.config:
            error(
                "<COR52619266E>", KeyError("Missing `module_paths` in workflow config!")
            )

        if module_paths_key not in self.config.module_paths:
            error(
                "<COR75976687E>",
                KeyError(
                    "Missing required {} key in config.module_paths!".format(
                        module_paths_key
                    )
                ),
            )

        module_list = self.config.module_paths[module_paths_key]
        error.type_check("<COR21790391E>", list, module_list=module_list)

        # Iterate through the list and load module one by one
        loaded_modules = []
        for module in module_list:
            module_path = os.path.join(self.model_path, module)
            loaded_modules.append(core.load(module_path))

        return loaded_modules



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Some type conversion helpers for going between python and protocol buffer interfaces
"""
# Standard
from inspect import isclass
from typing import Optional, Type, Union, get_args, get_origin

# Local
from caikit.core import ModuleBase
from caikit.core.data_model import DataBase, DataStream


def has_data_stream(arg_type: Type) -> bool:
    """Recursive check for a DataStream container in a type annotation"""
    if _is_data_stream(arg_type):
        return True

    typing_args = get_args(arg_type)
    if len(typing_args) > 0:
        for typ in typing_args:
            if has_data_stream(typ):
                return True

    return False


def get_data_stream_type(arg_type: Type) -> Optional[Type]:
    """Extracts the `DataStream` annotation from inside a nested type annotation, if it exists"""
    if _is_data_stream(arg_type):
        return arg_type
    typing_args = get_args(arg_type)
    for typ in typing_args:
        if has_data_stream(typ):
            return get_data_stream_type(typ)

    return None


def is_model_type(arg_type: Type) -> bool:
    if isclass(arg_type) and issubclass(arg_type, ModuleBase):
        return True
    typing_origin = get_origin(arg_type)
    typing_args = get_args(arg_type)

    if typing_origin is Union:
        for typ in typing_args:
            if isclass(typ) and issubclass(typ, ModuleBase):
                return True
    return False


def is_data_model_type(arg_type: Type) -> bool:
    return isclass(arg_type) and issubclass(arg_type, DataBase)


def _is_data_stream(arg_type: Type) -> bool:
    return arg_type == DataStream or get_origin(arg_type) == DataStream



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""DataStream validator that validates that a datastream includes the expected
parts - for dictionaries, that all expected dictionary keys exist, and for
iterables, that the iterable length is the expected length.
The types of the internal elements of each item are also validated.
"""
# Standard
from typing import Dict

# First Party
import alog

# Local
from ...exceptions import error_handler
from ...exceptions.validation_error import DataValidationError
from .. import DataStream

log = alog.use_channel("DATSTRMVLDTR")
error = error_handler.get(log)


class DataStreamValidator:
    """Validates that DataStreams contain the expected data items based on type"""

    def __init__(self, expected_keys: Dict[str, type]):
        """Initialize DataStreamValidator

        Args:
            expected_keys (Dict(str)): (Ordered) dictionary of key names ->
                types that determines how data will be validated
        """
        self._expected_keys: Dict[str, type] = expected_keys
        error.value_check(
            "<COR53878661E>", self._expected_keys, "`expected_keys` should be nonempty"
        )

    def validate(self, stream: DataStream) -> DataStream:
        """Attempt to validate the data items in a data stream (lazily)

        Validation checks:
        1. That the data in the stream is either a dictionary or list or tuple
        2. In the case of dictionaries, all expected keys exist and the values are the types
        given in
            `self._expected_keys`
        3. In the case of iterables, that the length of each iterable matches the expected number of
            items and the type of the nth item matches the nth type in `self._expected_keys`

        Args:
            stream (DataStream): stream intended to be converted

        Returns:
            The same data stream, which will now throw DataValidationErrors when accessed if the
            data is not valid.
        """
        data_item_number = -1

        def validate(data_item):
            # Enclose the counter here for later mutation when this map function is applied
            nonlocal data_item_number
            data_item_number += 1
            self._validate_data(data_item, data_item_number)
            return data_item

        return stream.map(validate)

    def _validate_data(self, data_item: object, data_item_number: int) -> None:
        """Validate a single data item from a data stream

        Args:
            data_item: A data object yielded by the stream
            data_item_number: The index of the object in the stream
        """
        if isinstance(data_item, dict):
            # From dictionary: error if key doesn't exist
            missing_keys = set(self._expected_keys.keys()) - set(data_item.keys())
            if len(missing_keys) > 0:
                message = "Data item is missing required key(s): {} ".format(
                    missing_keys
                )
                raise DataValidationError(message, data_item_number)
            # Error if any key is wrong type
            for key, type_ in self._expected_keys.items():
                if not isinstance(data_item[key], type_):
                    message = (
                        "Expected {} in data item to be of type {}, but got {}".format(
                            key, type_, type(data_item[key])
                        )
                    )
                    raise DataValidationError(message, data_item_number)

        elif isinstance(data_item, (list, tuple)):
            # From iterable: error on too many items, too few items
            if len(self._expected_keys) != len(data_item):
                message = "Expected data item to have {} elements but contained {} elements".format(
                    len(self._expected_keys), len(data_item)
                )
                raise DataValidationError(message, data_item_number)
            # Error if any element has wrong type
            for type_, element, index in zip(
                self._expected_keys.values(), data_item, range(len(data_item))
            ):
                if not isinstance(element, type_):
                    # pylint: disable=too-many-format-args
                    message = "Expected element {} in data item to be of type {}, "
                    "but got {}".format(index, type_, type(element))
                    raise DataValidationError(message, data_item_number)

        else:
            message = "This data item of type {} cannot be converted :(".format(
                type(data_item)
            )
            log.error("<COR76665827E>", message)

            raise DataValidationError(message, data_item_number)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The model management abstractions manage the key lifecycle phases of a concrete
model instance and are used by the ModelManager to handle each phase of the
model lifecycle.

For architectural details, see docs/adrs/020-model-management-abstractions.md
"""

# Local
from .factories import (
    model_finder_factory,
    model_initializer_factory,
    model_trainer_factory,
)
from .model_finder_base import ModelFinderBase
from .model_initializer_base import ModelInitializerBase
from .model_trainer_base import ModelTrainerBase, TrainingInfo



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""DataStream resolver that resolves a given file or Datastream to a DataStream
"""
# Standard
from typing import Dict

# First Party
import alog

# Local
from ...exceptions import error_handler
from .. import DataStream
from .converter import DataStreamConverter
from .csv_column_formatter import CSVColumnFormatter
from .validator import DataStreamValidator

log = alog.use_channel("DATSTRMRSLVR")
error = error_handler.get(log)


class DataStreamResolver:
    """Resolves files or DataStreams into DataStreams"""

    def __init__(self, target_stream_type: type, expected_keys: Dict[str, type]):
        """Initialize DataStreamResolver

        Args:
            target_stream_type (type): The target type for the data items in the
                resolved stream. dict and list are supported.
            expected_keys (dict(str, type)): Dictionary of keys -> types that
                determines how data will be formatted in the data stream.

                If you want a stream of dictionaries, we'll try to put these as keys in each one.
                If you want a stream of lists, we'll try to locate these values and place them in
                each list in order.

                In either case, we'll validate that each piece of data is of the type specified.

                See the DataStreamConverter docs for more info.
        """
        error.type_check("<COR24496441E>", type, target_stream_type=target_stream_type)
        error.type_check("<COR24496442E>", dict, expected_keys=expected_keys)

        self._converter: DataStreamConverter = DataStreamConverter(
            target_stream_type, list(expected_keys.keys())
        )
        self._validator: DataStreamValidator = DataStreamValidator(expected_keys)
        self._csv_formatter: CSVColumnFormatter = CSVColumnFormatter(expected_keys)

    def as_data_stream(self, file_or_data_stream) -> DataStream:
        """Marshals whatever you give it into a data stream, or dies trying.
        ...Or leaves the error for you to find later when reading the stream.

        Args:
            file_or_data_stream (str or DataStream): Either a string path to a
                file, or a DataStream
        Returns:
            DataStream: The data as a converted and properly formatted data stream
        """
        error.type_check(
            "<COR92088414E>", str, DataStream, file_or_data_stream=file_or_data_stream
        )

        if isinstance(file_or_data_stream, str):
            # File here
            loaded_stream = DataStream.from_file(file_or_data_stream)
            listified_stream = self._csv_formatter.format(loaded_stream)
            validated_stream = self._validator.validate(listified_stream)
            return self._converter.convert(validated_stream)

        # Datastream here
        listified_stream = self._csv_formatter.format(file_or_data_stream)
        validated_stream = self._validator.validate(listified_stream)
        return self._converter.convert(validated_stream)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Standard
from enum import Enum
import inspect

# First Party
# First party
import alog

# Local
from ..exceptions import error_handler

log = alog.use_channel("WIPDC")
error = error_handler.get(log)

################ Constants ##########################################

message_format = "{} is still in the {} phase and subject to change!"

_ENABLE_DECORATOR = True


class WipCategory(Enum):
    WIP = 1
    BETA = 2


class Action(Enum):
    ERROR = 1
    WARNING = 2


################ Implementation #####################################


def disable_wip():
    """Utility function to disable decorator functionality.
    Mainly designed for testing
    """
    # pylint: disable=global-statement
    global _ENABLE_DECORATOR
    _ENABLE_DECORATOR = False


def enable_wip():
    """Utility function to enable decorator functionality.
    Mainly designed for testing
    """
    # pylint: disable=global-statement
    global _ENABLE_DECORATOR
    _ENABLE_DECORATOR = True


class TempDisableWIP:
    """Temporarily disable wip decorator for a particular block of code

    NOTE: There is a potential race condition possible here in cases where
    other code using wip decorator gets called at the same time this context
    based disabling functionality is invoked. If this happens, the decorator will
    get disabled for all the functions invoking at the same time. This is
    because we are using the global disable / enable functions for this class.
    """

    ## Notes to fix:
    # As per current design of the decorator the enable / disable can only happen
    # globally at the invocation time. May be there is a way to overcome this by
    # changing the _get_message function dynamically.

    def __enter__(self):
        disable_wip()

    def __exit__(self, *args):
        enable_wip()


def work_in_progress(*args, **kwargs):
    """Decorator that can be used to mark a function
    or a class as "work in progress". It will result in a warning being emitted
    when the function / class is used.

    Args:
        category (WipCategory): Enum specifying what category of message you
            want to throw
        action (Action): Enum specifying what type of action you want to take.
            Example: ERROR or WARNING

    Example Usage:

    ### Decorating class

    1. No configuration:
        @work_in_progress
        class Foo:
            pass

    2. Action and category configuration:
        @work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
        class Foo:
            pass

    ### Decorating Function:

    1. No configuration:
        @work_in_progress
        def foo(*args, **kwargs):
            pass

    2. Action and category configuration:
        @work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
         def foo(*args, **kwargs):
            pass

    ### Sample message:

    foo is still in the BETA phase and subject to change!

    """
    if args:
        wrapped_obj = args[0]
    else:
        wrapped_obj = None

    # Set defaults
    category = kwargs.get("category", WipCategory.WIP)
    action = kwargs.get("action", Action.WARNING)

    # Type checks
    error.type_check("<TRU92076783E>", WipCategory, category=category)
    error.type_check("<TRU87572165E>", Action, action=action)

    if inspect.isclass(wrapped_obj) or inspect.isfunction(wrapped_obj):
        # TODO: if a class, add this decorator to all the functions of this class
        return _decorator_handler(wrapped_obj, category, action)

    if len(kwargs) > 0:

        def decorator(wrapped_obj):
            return _decorator_handler(wrapped_obj, category, action)

        return decorator

    raise ValueError(
        "Invalid usage of wip decorator. {} argument not supported!".format(
            type(wrapped_obj)
        )
    )


def _decorator_handler(wrapped_obj, category, action):
    """Utility function to cover common decorator handling
    logic.
    Args:
        wrapped_obj (Callable): Class or function to be decorated
        category (Enum(WipCategory)): Enum specifying the category of the
            message
        Action (Enum(Action)): Enum specifying the action to be taken with the
            decorator
    Returns:
        function:
            Decorator function
    """

    if inspect.isclass(wrapped_obj):
        # Replace __new__ function of wrapped class
        # with wrapped_cls function that includes
        # warning message
        new_class = wrapped_obj.__new__

        def wrapped_cls(cls, *args, **kwargs):
            _get_message(wrapped_obj, category, action)

            # if class __new__ is empty
            if new_class is object.__new__:
                return new_class(cls)

            return new_class(cls, *args, **kwargs)

        wrapped_obj.__new__ = staticmethod(wrapped_cls)

        return wrapped_obj

    def wip_decorator(*args, **kwargs):
        # function specific handling
        _get_message(wrapped_obj, category, action)
        return wrapped_obj(*args, **kwargs)

    return wip_decorator


def _get_message(wrapped_obj, category, action):
    """Utility function to run action"""
    if _ENABLE_DECORATOR:
        message = message_format.format(wrapped_obj, category.name)
        if action == Action.ERROR:
            raise RuntimeError(message)
        if action == Action.WARNING:
            log.warning(message)



# Standard
from collections import Counter
import json
import os
import threading
import time
import uuid

# First Party
import alog

# Local
from caikit import get_config

log = alog.use_channel("RPC_METER")


# pylint: disable=too-many-instance-attributes
class RPCMeter:
    """This class contains metering logic for RPC calls affiliated with the Caikit Runtime that are
    not a part of the Model Runtime proto definition.
    """

    def __init__(self):
        """Initialize a RPCMeter instance and start the log writer thread."""

        self.predict_rpc_counter = []
        self.logs_writer_thread = threading.Thread(target=self.write_metrics)
        self.metering_event = threading.Event()
        self.rpc_counter_lock = threading.Lock()
        self.write_file_lock = threading.Lock()
        self.metrics_dir = get_config().runtime.metering.log_dir
        self.unique_id = str(uuid.uuid4()).replace("-", "_")
        self.file_path = os.path.join(
            self.metrics_dir,
            "predict_rpc_metrics_{}.json".format(self.unique_id),
        )
        if not os.path.exists(self.metrics_dir):
            os.makedirs(self.metrics_dir)
        self.logs_writer_thread.start()
        log.debug(
            "<RUN76774000I>",
            "Started metering log writer thread %s",
            self.logs_writer_thread.name,
        )

    def update_metrics(self, model_type):
        """Updates metrics, writes to file if max count has reached and resets counters
        Args:
            model_type (string): Type of model the request was made for
        """
        # Locking to ensure that with concurrent updates to counters, the latest metrics are
        # reported
        with self.rpc_counter_lock:
            self.predict_rpc_counter.append(model_type)

    def end_writer_thread(self):
        """Kills log writer thread"""
        self.metering_event.set()
        if self.logs_writer_thread.is_alive():
            self.logs_writer_thread.join()

    def flush_metrics(self):
        """Writes metrics and kills log writer thread"""
        log.debug(
            "<RUN76774001I>",
            "Server interrupted so flushing metrics to file for thread %s",
            self.logs_writer_thread.name,
        )
        self.end_writer_thread()
        self._write_metrics()

    def write_metrics(self):
        """Function for log writer thread to write logs at specific intervals configured by user"""
        while True:
            log.debug(
                "<RUN76774002I>",
                "Metering log file writing to %s",
                self.file_path,
            )
            self._write_metrics()
            notified = self.metering_event.wait(
                get_config().runtime.metering.log_interval
            )
            if notified:
                log.debug("<RUN76774003I>", "Shutting down metering writer log thread")
                break

    def _write_metrics(self):
        """Writes all metrics to directory specified in config and resets counters"""
        try:
            metrics_dict = {}
            with self.rpc_counter_lock:
                if self.predict_rpc_counter:
                    metrics_dict = {
                        "timestamp": time.time(),
                        "batch_size": len(self.predict_rpc_counter),
                        "model_type_counters": Counter(self.predict_rpc_counter),
                        "container_id": self.unique_id,
                    }
                    self.predict_rpc_counter.clear()
                    log.debug("<RUN76774004I>", "predict_rpc_counter reset")
            if metrics_dict:
                with self.write_file_lock:
                    json_string = json.dumps(metrics_dict)
                    with open(self.file_path, "a", encoding="utf-8") as json_file:
                        json_file.write(json_string + "\n")
                # Log the metrics dict to stdout as well for later scraping if required
                log.info("<RUN76774008I>", metrics_dict)
                log.info(
                    "<RUN76774005I>",
                    "Successfully written metrics file to %s",
                    self.file_path,
                )
            else:
                log.info(
                    "<RUN76774006I>", "No new RPCs to write, skipping metering logging"
                )
        except json.JSONDecodeError as e:
            log.info("<RUN76774007I>", "Write metrics failed with %s", str(e))
        except FileNotFoundError as e:
            log.info("<RUN76774007I>", "Write metrics failed with %s", str(e))

    def __del__(self):
        self.end_writer_thread()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobufs-defined services."""
# Third Party
import grpc

# Local
from . import model_runtime_pb2 as model__runtime__pb2


class ModelRuntimeStub(object):
    """this is the internal "sidecar" API for interfacing with a
    colocated model runtime container
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.loadModel = channel.unary_unary(
            "/mmesh.ModelRuntime/loadModel",
            request_serializer=model__runtime__pb2.LoadModelRequest.SerializeToString,
            response_deserializer=model__runtime__pb2.LoadModelResponse.FromString,
        )
        self.unloadModel = channel.unary_unary(
            "/mmesh.ModelRuntime/unloadModel",
            request_serializer=model__runtime__pb2.UnloadModelRequest.SerializeToString,
            response_deserializer=model__runtime__pb2.UnloadModelResponse.FromString,
        )
        self.predictModelSize = channel.unary_unary(
            "/mmesh.ModelRuntime/predictModelSize",
            request_serializer=model__runtime__pb2.PredictModelSizeRequest.SerializeToString,
            response_deserializer=model__runtime__pb2.PredictModelSizeResponse.FromString,
        )
        self.modelSize = channel.unary_unary(
            "/mmesh.ModelRuntime/modelSize",
            request_serializer=model__runtime__pb2.ModelSizeRequest.SerializeToString,
            response_deserializer=model__runtime__pb2.ModelSizeResponse.FromString,
        )
        self.runtimeStatus = channel.unary_unary(
            "/mmesh.ModelRuntime/runtimeStatus",
            request_serializer=model__runtime__pb2.RuntimeStatusRequest.SerializeToString,
            response_deserializer=model__runtime__pb2.RuntimeStatusResponse.FromString,
        )


class ModelRuntimeServicer(object):
    """this is the internal "sidecar" API for interfacing with a
    colocated model runtime container
    """

    def loadModel(self, request, context):
        """Load a model, return when model is fully loaded.
        Include size of loaded model in response if no additional cost.
        A gRPC error code of PRECONDITION_FAILED or INVALID_ARGUMENT
        should be returned if no attempt to load the model was made
        (so can be sure that no space remains used).
        Note that the RPC may be cancelled by model-mesh prior to completion,
        after which an unloadModel call will immediately be sent for the same model.
        To avoid state inconsistency and "leaking" memory, implementors should
        ensure that this case is properly handled, i.e. that the model doesn't
        remain loaded after returning successfully from this unloadModel call.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def unloadModel(self, request, context):
        """Unload a previously loaded (or failed) model. Return when model
        is fully unloaded, or immediately if not found/loaded.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def predictModelSize(self, request, context):
        """Predict size of not-yet-loaded model - must return almost immediately.
        Should not perform expensive computation or remote lookups.
        Should be a conservative estimate.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def modelSize(self, request, context):
        """Calculate size (memory consumption) of currently-loaded model"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def runtimeStatus(self, request, context):
        """Provide basic runtime status and parameters; called only during startup.
        Before returning a READY status, implementations should check for and
        purge any/all currently-loaded models. Since this is only called during
        startup, there should very rarely be any, but if there are it implies
        the model-mesh container restarted unexpectedly and such a purge must
        be done to ensure continued consistency of state and avoid over-committing
        resources.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ModelRuntimeServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "loadModel": grpc.unary_unary_rpc_method_handler(
            servicer.loadModel,
            request_deserializer=model__runtime__pb2.LoadModelRequest.FromString,
            response_serializer=model__runtime__pb2.LoadModelResponse.SerializeToString,
        ),
        "unloadModel": grpc.unary_unary_rpc_method_handler(
            servicer.unloadModel,
            request_deserializer=model__runtime__pb2.UnloadModelRequest.FromString,
            response_serializer=model__runtime__pb2.UnloadModelResponse.SerializeToString,
        ),
        "predictModelSize": grpc.unary_unary_rpc_method_handler(
            servicer.predictModelSize,
            request_deserializer=model__runtime__pb2.PredictModelSizeRequest.FromString,
            response_serializer=model__runtime__pb2.PredictModelSizeResponse.SerializeToString,
        ),
        "modelSize": grpc.unary_unary_rpc_method_handler(
            servicer.modelSize,
            request_deserializer=model__runtime__pb2.ModelSizeRequest.FromString,
            response_serializer=model__runtime__pb2.ModelSizeResponse.SerializeToString,
        ),
        "runtimeStatus": grpc.unary_unary_rpc_method_handler(
            servicer.runtimeStatus,
            request_deserializer=model__runtime__pb2.RuntimeStatusRequest.FromString,
            response_serializer=model__runtime__pb2.RuntimeStatusResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "mmesh.ModelRuntime", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class ModelRuntime(object):
    """this is the internal "sidecar" API for interfacing with a
    colocated model runtime container
    """

    @staticmethod
    def loadModel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelRuntime/loadModel",
            model__runtime__pb2.LoadModelRequest.SerializeToString,
            model__runtime__pb2.LoadModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def unloadModel(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelRuntime/unloadModel",
            model__runtime__pb2.UnloadModelRequest.SerializeToString,
            model__runtime__pb2.UnloadModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def predictModelSize(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelRuntime/predictModelSize",
            model__runtime__pb2.PredictModelSizeRequest.SerializeToString,
            model__runtime__pb2.PredictModelSizeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def modelSize(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelRuntime/modelSize",
            model__runtime__pb2.ModelSizeRequest.SerializeToString,
            model__runtime__pb2.ModelSizeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def runtimeStatus(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mmesh.ModelRuntime/runtimeStatus",
            model__runtime__pb2.RuntimeStatusRequest.SerializeToString,
            model__runtime__pb2.RuntimeStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data structures for text generation representations"""

# Standard
from enum import Enum
from typing import List, Optional

# Third Party
import numpy as np

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber
import alog

# Local
from ....core import DataObjectBase, dataobject
from ...common.data_model import ProducerId
from .package import NLP_PACKAGE

log = alog.use_channel("DATAM")


@dataobject(package=NLP_PACKAGE)
class FinishReason(Enum):
    NOT_FINISHED = 0
    MAX_TOKENS = 1
    EOS_TOKEN = 2
    CANCELLED = 3
    TIME_LIMIT = 4
    STOP_SEQUENCE = 5
    TOKEN_LIMIT = 6
    ERROR = 7


@dataobject(package=NLP_PACKAGE)
class GeneratedTextResult(DataObjectBase):
    generated_text: Annotated[str, FieldNumber(1)]
    generated_tokens: Annotated[int, FieldNumber(2)]
    finish_reason: Annotated[FinishReason, FieldNumber(3)]
    producer_id: Annotated[ProducerId, FieldNumber(4)]
    input_token_count: Annotated[int, FieldNumber(5)]
    seed: Annotated[Optional[np.uint64], FieldNumber(6)]


@dataobject(package=NLP_PACKAGE)
class GeneratedToken(DataObjectBase):
    text: Annotated[str, FieldNumber(1)]
    logprob: Annotated[Optional[float], FieldNumber(3)]


@dataobject(package=NLP_PACKAGE)
class TokenStreamDetails(DataObjectBase):
    finish_reason: Annotated[FinishReason, FieldNumber(1)]
    generated_tokens: Annotated[np.uint32, FieldNumber(2)]
    seed: Annotated[Optional[np.uint64], FieldNumber(3)]
    input_token_count: Annotated[int, FieldNumber(4)]


@dataobject(package=NLP_PACKAGE)
class GeneratedTextStreamResult(DataObjectBase):
    generated_text: Annotated[str, FieldNumber(1)]
    tokens: Annotated[Optional[List[GeneratedToken]], FieldNumber(2)]
    details: Annotated[Optional[TokenStreamDetails], FieldNumber(3)]
    producer_id: Annotated[ProducerId, FieldNumber(4)]



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import Callable
import abc
import threading

# First Party
import alog

# Local
from caikit.core.toolkit.destroyable_thread import DestroyableThread
from caikit.runtime.types.aborted_exception import AbortedException

log = alog.use_channel("ABORT-ACTION")


class ActionAborter(abc.ABC):
    """Simple interface to wrap up a notification that an action must abort.

    Children of this class can bind to any notification tool (e.g. grpc context)
    """

    @abc.abstractmethod
    def must_abort(self) -> bool:
        """Indicate whether or not the action must be aborted"""

    @abc.abstractmethod
    def add_event(self, event: threading.Event):
        """Add an event to notify when abort happens"""


class AbortableAction:
    """A class for Abortable Actions. We want actions that are computationally heavy to be
    abortable by Model Mesh! Currently, we use this for the following operations.

    - Loading a model
    - Predicting with a model
    - Training a model

    In the future, this may include getting the size of a model, depending on how that we choose
    to implement that.

    How it works:
        Instances of this class create a threading.Event, which will be used to signal that either:
        - The RPC was terminated
        - The heavy work that we wanted to complete is done
        This is done by using a RpcAborter and a DestroyableThread.
        Registering the event with the RpcAborter will cause it to set when the RPC is
        terminated, and creating a DestroyableThread with the event will cause it to set when
        the thread terminates.

        The action will start the DestroyableThread and then wait on the event. When it wakes, it
        will check the reason and destroy the thread if it was woken by the RpcAborter or return
        the result if it was woken by the thread completing.
    """

    def __init__(
        self,
        call_aborter: ActionAborter,
        runnable_func: Callable,
        *args,
        **kwargs,
    ):
        """
        Args:
            call_aborter - call aborter capable of aborting the runnable_func
            runnable_func - the function to be run as an abortable action
            *args - nonkeyword arguments to runnable_func
            **kwargs - keyword arguments to runnable_func"""

        # Create new event to watch for both RPC termination and work completion
        self.__done_or_aborted_event = threading.Event()

        # Register the event with our call aborter so it fires if the RPC terminates
        self.call_aborter = call_aborter
        self.call_aborter.add_event(self.__done_or_aborted_event)

        # Create a new thread to do the work, which will set the event if it finished
        self.__runnable_func = runnable_func
        self.__work_thread = DestroyableThread(
            self.__runnable_func,
            *args,
            work_done_event=self.__done_or_aborted_event,
            **kwargs,
        )

    def do(self):
        # Start the work and wait
        self.__work_thread.start()
        self.__done_or_aborted_event.wait()

        # Now, check the call aborter to see what happened.
        # Option 1: The RPC was terminated. Kill the work thread and raise an exception
        if self.call_aborter.must_abort():
            log.info(
                "<RUN14653271I>", "Aborting work in progress: %s", self.__runnable_func
            )
            self.__work_thread.destroy()
            self.__work_thread.join()
            raise AbortedException("Aborted work: {}".format(self.__runnable_func))

        # Options 2: Work thread finished normally. Hooray!
        log.debug("Work finished: %s", self.__runnable_func)
        self.__work_thread.join()
        return self.__work_thread.get_or_throw()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A generic module to help Predict and Train servicers
"""
# Standard
from functools import update_wrapper
from typing import Any, Dict, Iterable, Iterator, Tuple
import sys
import traceback

# Third Party
from google.protobuf.descriptor import FieldDescriptor, ServiceDescriptor
from google.protobuf.message import Message as ProtoMessageType
import grpc

# First Party
import alog

# Local
from caikit.core.data_model import DataStream
from caikit.core.data_model.base import DataBase
from caikit.core.exceptions.caikit_core_exception import (
    CaikitCoreException,
    CaikitCoreStatusCode,
)
from caikit.core.signature_parsing import CaikitMethodSignature
from caikit.interfaces.runtime.data_model.training_management import ModelPointer
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.service_generation.data_stream_source import (
    DataStreamSourceBase,
    make_data_stream_source,
)
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.utils.import_util import get_data_model

log = alog.use_channel("SERVICR-UTIL")

# Protobuf non primitives
# Ref: https://developers.google.com/protocol-buffers/docs/reference/cpp/google.protobuf.descriptor
NON_PRIMITIVE_TYPES = [FieldDescriptor.TYPE_MESSAGE, FieldDescriptor.TYPE_ENUM]


def validate_caikit_library_class_exists(cdm, class_name):
    try:
        # Attempt to instantiate a Caikit* CDM class of the specified
        # output type
        return getattr(cdm, class_name)

    except AttributeError as e:
        log.warning("<RUN24024050W>", "No Caikit Library CDM class for %s", class_name)
        # Look up the data model class corresponding to the given name
        dm_class_name = DataBase.get_class_for_name(class_name=class_name)
        if not dm_class_name:
            log.error(
                "<RUN24024010E>", "No Caikit Library CDM class for %s", class_name
            )
            raise e
        return dm_class_name


def validate_caikit_library_class_method_exists(caikit_library_class, method_name):
    try:
        getattr(caikit_library_class, method_name)

    except AttributeError as e:
        log.error(
            "<RUN24024009E>",
            "No `%s` method for Caikit Library CDM class %s",
            method_name,
            caikit_library_class,
        )
        raise e


def build_proto_stream(
    caikit_library_response: Iterable[DataBase],
) -> Iterator[ProtoMessageType]:
    """Returns an iterator that serializes each item in the model's response to protobuf"""

    def _proto_generator():
        for item in caikit_library_response:
            try:
                yield item.to_proto()
            except Exception as e:
                log.warning(
                    {
                        "log_code": "<RUN11567943W>",
                        "message": "Exception while serializing response from stream: "
                        "{}".format(e),
                        "stack_trace": traceback.format_exc(),
                    }
                )
                raise CaikitRuntimeException(
                    grpc.StatusCode.INTERNAL,
                    "Could not serialize output in model response stream",
                ) from e

    return iter(DataStream(_proto_generator))


def build_proto_response(caikit_library_response: DataBase) -> ProtoMessageType:
    """Serializes a data model instance into a protobuf message"""
    try:
        return caikit_library_response.to_proto()
    except Exception as e:
        log.warning(
            {
                "log_code": "<RUN11230943W>",
                "message": "Exception while serializing response: {}".format(e),
                "stack_trace": traceback.format_exc(),
            }
        )
        raise CaikitRuntimeException(
            grpc.StatusCode.INTERNAL, "Could not serialize response"
        ) from e


def is_protobuf_primitive_field(obj):
    """Check whether or not a descriptor object is a protobufs primitive. The
    full list of descriptors can be found at the link below.

    https://developers.google.com/protocol-buffers/docs/reference/cpp/google.protobuf.descriptor

    Args:
        obj (google.protobufs.pyext._message.FieldDescriptor):
            A protoc-compiled FieldDescriptor object describing a field to be
            passed as input or output in a well-structured way.
    Returns:
        boolean indicating whether or not a primitive compiled protobuf field
        was passed.
    """
    # If we have a non-primitive object
    if not isinstance(obj, FieldDescriptor):
        log.warning(
            "<RUN24033310D>",
            "Type [%s] should be FieldDescriptor",
            type(obj),
        )
        return False

    if obj.type == FieldDescriptor.TYPE_MESSAGE and obj.message_type.name == "Struct":
        return True
    return obj.type not in NON_PRIMITIVE_TYPES


def get_metadata(context, key, required=True):
    """Retrieve a value from the gRPC ServicerContext invocation metadata
    dictionary with the given key

    Args:
        context(grpc.ServicerContext): Context object (contains request metadata, etc)
        key(string): The invocation metadata dictionary key to retrieve
        required(boolean): Whether the piece of metadata is required to exist
    Returns:
        value (object): The value, or None if the specified key does not
            exist in the invocation metadata
    """
    try:
        return dict(context.invocation_metadata()).get(key, None)
    except Exception as e:  # pylint: disable=broad-exception-caught
        log.warning(
            {
                "log_code": "<RUN15349983W>",
                "message": "Exception while attempting to extract {} from "
                "request metadata: {}".format(key, e),
                "stack_trace": traceback.format_exc(),
            }
        )
        if required:
            raise CaikitRuntimeException(
                grpc.StatusCode.INVALID_ARGUMENT,
                "Could not read required metadata: {}".format(key),
            ) from e


def validate_data_model(
    service_descriptor: ServiceDescriptor,
):
    """Validate the Caikit Library Common Data Model (CDM) against a service descriptor
    that defines the RPCs that this class must support at predict/train time.
    More specifically, this function will iterate over every RPC/method
    defined in the service descriptor, and for each RPC, will validate that
    the Caikit Library CDM has classes that correspond with the field types of the
    incoming message fields, and that those Caikit Library classes have `from_proto`
    functions defined.  It will further verify that the output message
    has a corresponding Caikit Library CDM class with a `to_proto` method defined.

    Args:
        service_descriptor(google.protobuf.descriptor.ServiceDescriptor):
            A protoc-compiled ServiceDescriptor that defines the predict
            RPCs that will be serviced by this GlobalPredictServicer

    Raises:
        AttributeError if a discrepancy is found between the RPC service
        descriptor and the Caikit Library CDM, which will prevent an instance of
        this class from being instantiated
    """
    # this `cdm` was moved here from import-time
    cdm = get_data_model()
    for method in service_descriptor.methods:
        # Retrieve the descriptor of the input message for this RPC, and
        # verify that each field of the input message can be translated
        # into a corresponding object of the Caikit Library CDM, and that each
        # Caikit Library CDM class has a `from_proto` method defined
        input_proto_msg = method.input_type

        # Iterate over each field in this input RPC message...
        for field in input_proto_msg.fields:
            # and check to make that it is either a primitive protobufs type or that
            # we have a data model class that we can deserialize the protobufs with
            if not is_protobuf_primitive_field(field):

                if field.message_type and field.message_type.GetOptions().map_entry:
                    log.debug(
                        "<RUN51658878D>",
                        "Field: [%s] is a map of key type: [%s] and value type: [%s]",
                        field.name,
                        field.message_type.fields[0].type,
                        field.message_type.fields[1].type,
                    )
                    continue

                # ... or that we can get the field type name, e.g., RawDocument...
                field_type = input_proto_msg.fields_by_name[
                    field.name
                ].message_type.name

                # ...and ensuring that we can load a corresponding object from the Caikit* CDM
                caikit_library_class = validate_caikit_library_class_exists(
                    cdm, field_type
                )

                # ...and also ensuring that the Caikit Library CDM class has a `from_proto`
                # method...
                validate_caikit_library_class_method_exists(
                    caikit_library_class, "from_proto"
                )
            else:
                log.debug(
                    "<RUN51658879D>",
                    "Field: [%s] is a primitive of type: [%s]",
                    field.name,
                    field.type,
                )

        # Now we do something similar for the output RPC message, verifying
        # that we can construct an object of the Caikit Library CDM that matches
        # the specified field type, and that said Caikit Library object has a
        # to_proto method defined. No need to check for proto primitives here since
        # all Caikit library modules should return well formed "predict" messages
        # from the data model.
        output_class = method.output_type.name
        caikit_Library_class = validate_caikit_library_class_exists(cdm, output_class)
        validate_caikit_library_class_method_exists(caikit_Library_class, "to_proto")


class ServicePackageStreamWrapper(DataStreamSourceBase):
    """This class wraps up a DataStreamSourceBase derived class so that it can
    be safely pickled to a subprocess using "spawn".

    When using "spawn" to start a subprocess, a clean python process is created
    and none of the runtime code leading to the creation is re-executed. For
    most classes that are created at import time, this is fine, but for
    dynamically created classes like those created with make_data_stream_source,
    the class is not recreated on spawn and cannot be looked up.

    To get around this, when unpickling this class, it falls back to performing
    service generation dynamically if the class doesn't already exist
    """

    def __init__(self, stream: DataStreamSourceBase):
        """Initialize with the data stream source to wrap

        NOTE: We _intentionally_ don't initialize the base class here and
            instead we forward
        """
        self._stream = stream
        super().__init__()
        update_wrapper(self, stream)

    def __getstate__(self) -> Tuple[str, str, type, bytes]:
        """Pickle as a tuple of primitive types

        Returns:
            stream_module_name (str): The name of the module that holds the
                stream source class
            stream_class_name (str): The name of the stream source class itself
            stream_class_element_type (type): The type of the elements for this
                stream source
            content (bytes): The serialized bytes content of the underlying
                stream source as a dataobject
        """
        return (
            self._stream.__class__.__module__,
            self._stream.__class__.__name__,
            self._stream.__class__.ELEMENT_TYPE,
            self._stream.__getstate__(),
        )

    def __setstate__(self, pickled: Tuple[str, str, type, bytes]):
        """Unpickle the values from pickling and fall back to dynamic service
        generation if needed.

        Args:
            pickled (Tuple[str, str, type, bytes]): The four elements created
                when pickling this instance
        """
        stream_type_module, stream_type_name, element_type_class, stream_bytes = pickled
        try:
            mod = sys.modules[stream_type_module]
            stream_type = getattr(mod, stream_type_name)
        except (KeyError, AttributeError):
            # Local
            from caikit.runtime.service_factory import (  # pylint: disable=import-outside-toplevel
                ServicePackageFactory,
            )

            ServicePackageFactory.get_service_package(
                ServicePackageFactory.ServiceType.INFERENCE
            )
            ServicePackageFactory.get_service_package(
                ServicePackageFactory.ServiceType.TRAINING
            )
            stream_type = make_data_stream_source(element_type_class)
        self._stream = stream_type.__new__(stream_type)
        self._stream.__setstate__(stream_bytes)

    def __getattr__(self, name: str) -> Any:
        """Forward all getattr requests to the underlying stream"""
        return getattr(self._stream, name)

    def __setattr__(self, name: str, val: Any):
        """Forward attribute setting for generic attrs to the stream"""
        if name == "_stream":
            super().__setattr__(name, val)
        else:
            stream_attrs = list(self._stream.__class__.__annotations__.keys()) + list(
                self._stream.get_proto_class().DESCRIPTOR.fields_by_name.keys()
            )
            if name in stream_attrs:
                setattr(self._stream, name, val)
            else:
                super().__setattr__(name, val)


def build_caikit_library_request_dict(
    request: ProtoMessageType,
    module_signature: CaikitMethodSignature,
) -> Dict[str, Any]:
    """Build the request kwargs dict.

    Args:
        request (ProtoMessageType):
            The request proto message to deserialize from
        module_signature (CaikitMethodSignature):
            Module signature or metadata about method on a module

    Returns:
        kwargs dict
    """
    try:
        # Request messages are data model objects so .from_proto can be used
        request_data_model_class = DataBase.get_class_for_proto(request)
        request_data_model = request_data_model_class.from_proto(request)

        # Initialize kwargs from data model fields
        kwargs_dict = request_data_model.to_kwargs()

        # 1. Remove any fields not in request
        unset_field_names = []
        for field in request.DESCRIPTOR.fields:
            try:
                if not request.HasField(field.name):
                    unset_field_names.append(field.name)
            except ValueError as e:
                log.debug2(
                    "failed to check HasField on field %s, error: %s",
                    field.name,
                    e,
                )
                # Remove empty iterables since we cannot distinguish between
                # unset and empty repeated fields
                field_value = getattr(request, field.name)
                if isinstance(field_value, Iterable) and len(field_value) == 0:
                    unset_field_names.append(field.name)
        for unset_field_name in unset_field_names:
            if unset_field_name in kwargs_dict:
                kwargs_dict.pop(unset_field_name)

        # 2. Remove any fields not in the module signature
        absent_field_names = [
            field
            for field in kwargs_dict.keys()
            if field not in module_signature.parameters.keys()
        ]
        for absent_field_name in absent_field_names:
            kwargs_dict.pop(absent_field_name)

        # 3. Handle type conversions
        updated_kwargs = {}
        for field_name, field_value in kwargs_dict.items():

            # 3.1 Model Pointers
            if isinstance(field_value, ModelPointer):
                log.debug2("field %s value is a ModelPointer obj", field_name)
                model_manager = ModelManager.get_instance()
                model_retrieved = model_manager.retrieve_model(field_value.model_id)
                updated_kwargs[field_name] = model_retrieved

            # 3.2 Data streams
            elif isinstance(field_value, DataStreamSourceBase):
                log.debug2("field %s value is a DataStreamSourceBase", field_name)
                updated_kwargs[field_name] = ServicePackageStreamWrapper(field_value)

        # Apply the type conversion updates
        kwargs_dict.update(updated_kwargs)
        log.debug2("caikit_library_request_dict returned is: %s", kwargs_dict)
        return kwargs_dict

    except CaikitRuntimeException as e:
        log.warning(
            {
                "log_code": "<RUN50530381W>",
                "message": e.message,
                "error_id": e.id,
                **e.metadata,
            }
        )
        raise e

    except Exception as e:
        log.warning(
            {
                "log_code": "<RUN15438843W>",
                "message": "Exception while deserializing request: {}".format(e),
                "stack_trace": traceback.format_exc(),
            }
        )
        raise CaikitRuntimeException(
            grpc.StatusCode.INVALID_ARGUMENT, "Could not deserialize request"
        ) from e


def raise_caikit_runtime_exception(exception: CaikitCoreException):
    if exception.status_code == CaikitCoreStatusCode.NOT_FOUND:
        log.debug2("Not Found Error: [%s]", exception.message, exc_info=True)
        raise CaikitRuntimeException(
            grpc.StatusCode.NOT_FOUND,
            exception.message,
        ) from exception

    if exception.status_code == CaikitCoreStatusCode.INVALID_ARGUMENT:
        log.debug2("Invalid Argument Error: [%s]", exception.message, exc_info=True)
        raise CaikitRuntimeException(
            grpc.StatusCode.INVALID_ARGUMENT,
            exception.message,
        ) from exception

    if exception.status_code == CaikitCoreStatusCode.FORBIDDEN:
        log.debug2("Forbidden Error: [%s]", exception.message, exc_info=True)
        raise CaikitRuntimeException(
            grpc.StatusCode.PERMISSION_DENIED,
            exception.message,
        ) from exception

    if exception.status_code == CaikitCoreStatusCode.CONNECTION_ERROR:
        log.debug2("Connection Error: [%s]", exception.message, exc_info=True)
        raise CaikitRuntimeException(
            grpc.StatusCode.UNAVAILABLE,
            exception.message,
        ) from exception

    if exception.status_code == CaikitCoreStatusCode.UNAUTHORIZED:
        log.debug2("Unauthorized Error: [%s]", exception.Message, exc_info=True)
        raise CaikitRuntimeException(
            grpc.StatusCode.UNAUTHENTICATED,
            exception.message,
        ) from exception

    if exception.status_code == CaikitCoreStatusCode.FATAL:
        log.debug2("Fatal Error: [%s]", exception.message, exc_info=True)
        raise CaikitRuntimeException(
            grpc.StatusCode.INTERNAL,
            exception.message,
        ) from exception

    if exception.status_code == CaikitCoreStatusCode.UNKNOWN:
        log.debug2("Unknown Error: [%s]", exception.message, exc_info=True)
        raise CaikitRuntimeException(
            grpc.StatusCode.UNKNOWN,
            exception.message,
        ) from exception

    log.debug2(
        "Unhandled CaikitCoreException Error: [%s]", exception.message, exc_info=True
    )
    raise CaikitRuntimeException(
        grpc.StatusCode.INTERNAL, exception.message
    ) from exception



"""
A sample module for sample things!
"""
# Standard
from typing import Union

# Local
from ...data_model.sample import OtherOutputType, OtherTask, SampleInputType
from caikit.core.data_model import DataStream
from caikit.core.data_model.producer import ProducerId
from caikit.core.modules import ModuleLoader, ModuleSaver
import caikit.core


@caikit.core.module(
    "33221100-0405-0607-0809-0a0b02dd0e0f", "OtherModule", "0.0.1", OtherTask
)
class OtherModule(caikit.core.ModuleBase):
    def __init__(self, batch_size=64, learning_rate=0.0015):
        super().__init__()
        self.batch_size = batch_size
        self.learning_rate = learning_rate

    def run(
        self, sample_input: Union[SampleInputType, str], include_producer_id=True
    ) -> Union[OtherOutputType, str]:
        farewell = f"goodbye: {sample_input.name} {self.batch_size} times"
        producer_id = ProducerId("other_mod", "1.1.1") if include_producer_id else None
        return OtherOutputType(farewell, producer_id)

    @classmethod
    def load(cls, model_path, **kwargs):
        loader = ModuleLoader(model_path)
        config = loader.config
        return cls(config["train"]["batch_size"], config["train"]["learning_rate"])

    def save(self, model_path):
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            config_options = {
                "train": {
                    "batch_size": self.batch_size,
                    "learning_rate": self.learning_rate,
                },
            }

            module_saver.update_config(config_options)

    @classmethod
    def train(
        cls,
        training_data: DataStream[int],
        sample_input: Union[SampleInputType, str],
        batch_size: int = 64,
    ) -> "OtherModule":
        """Sample training method that produces a trained model"""
        assert isinstance(sample_input, SampleInputType) or isinstance(
            sample_input, str
        )
        # Barf if we were incorrectly passed data not in datastream format
        assert isinstance(training_data, DataStream)
        assert batch_size > 0
        return cls(batch_size=batch_size)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for the MultiModelFinder
"""
# Standard
from contextlib import contextmanager

# Third Party
import pytest

# First Party
import aconfig

# Local
from caikit.core import ModelManager
from caikit.core.model_management.multi_model_finder import MultiModelFinder
from tests.conftest import temp_config
from tests.core.helpers import TestFinder

## Helpers #####################################################################


@contextmanager
def temp_config_setup(
    multi_finder_name="default",
    multi_finder_cfg=None,
    config_overrides=None,
):
    config_overrides = config_overrides or {}
    finders = config_overrides.setdefault("model_management", {},).setdefault(
        "finders",
        {},
    )
    # Add a local one to proxy to
    if not finders:
        finders["local"] = {"type": "LOCAL"}
    multi_finder_cfg = multi_finder_cfg or {"finder_priority": ["local"]}

    multi_model_section = {
        "type": MultiModelFinder.name,
        "config": multi_finder_cfg,
    }
    finders.setdefault(multi_finder_name, multi_model_section)
    with temp_config(config_overrides, "merge"):
        config_overrides = aconfig.Config(config_overrides, override_env_vars=False)
        yield config_overrides.model_management.finders[multi_finder_name].config


@contextmanager
def temp_config_finder(
    multi_finder_name="default",
    multi_finder_cfg=None,
    config_overrides=None,
):
    with temp_config_setup(
        multi_finder_name=multi_finder_name,
        multi_finder_cfg=multi_finder_cfg,
        config_overrides=config_overrides,
    ) as mf_config:
        mmgr = ModelManager()
        mf_config.model_manager = mmgr
        yield MultiModelFinder(mf_config, multi_finder_name)


## Tests #######################################################################


def test_multi_model_finder_model_found(good_model_path):
    """Make sure that a simple proxy to local works"""
    with temp_config_finder() as finder:
        assert isinstance(finder, MultiModelFinder)
        assert finder.find_model(good_model_path)


@pytest.mark.parametrize(
    "test_finder_config",
    [{"fail_to_find": True}, {"raise_on_find": True}],
)
def test_multi_model_finder_first_not_found(test_finder_config, good_model_path):
    """Make sure that a model can be found if the first finder fails"""
    with temp_config_finder(
        config_overrides={
            "model_management": {
                "finders": {
                    "default": {
                        "type": MultiModelFinder.name,
                        "config": {
                            "finder_priority": ["test-not-found", "local"],
                        },
                    },
                    "test-not-found": {
                        "type": TestFinder.name,
                        "config": test_finder_config,
                    },
                    "local": {"type": "LOCAL"},
                }
            }
        }
    ) as finder:
        assert isinstance(finder, MultiModelFinder)
        assert finder.find_model(good_model_path)


def test_multi_model_finder_not_found():
    """Make sure that a simple proxy to local works"""
    with temp_config_finder() as finder:
        assert isinstance(finder, MultiModelFinder)
        assert finder.find_model("not/a/valid/path") is None


@pytest.mark.parametrize(
    "params",
    [
        ({"finder_priority": "local"}, TypeError),
        ({"finder_priority": [0, 1]}, TypeError),
        ({"finder_priority": []}, ValueError),
        ({"finder_priority": ["invalid-name"]}, ValueError),
        ({"finder_priority": ["default"]}, ValueError),
    ],
)
def test_multi_model_finder_invalid_config(params):
    """Validate all forms of bad config"""
    multi_finder_cfg, error_type = params
    with temp_config_setup(multi_finder_cfg=multi_finder_cfg) as mf_config:
        with pytest.raises(error_type):
            MultiModelFinder(mf_config, "default")



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A LoadedModel is a metadata wrapper around an instance of a core.ModuleBase
class that contains the additional information needed to manage that model in
the runtime.
"""

# Standard
from concurrent.futures import Future
from typing import Callable, Optional

# First Party
import alog

# Local
from caikit.core import ModuleBase
from caikit.core.exceptions import error_handler
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException

log = alog.use_channel("LOADED-MODEL")
error = error_handler.get(log)


# A future object that will yield an instance of a caikit module (a model)
# NOTE: 3.9 introduced subscript typing for Futures
try:
    CaikitModelFuture = Future[ModuleBase]  # pylint: disable=unsubscriptable-object
except TypeError:  # pragma: no cover
    CaikitModelFuture = Future

CaikitModelFutureFactory = Callable[[], CaikitModelFuture]


class LoadedModel:  # pylint: disable=too-many-instance-attributes
    __doc__ = __doc__

    class Builder:
        """The LoadedModel.Builder allows the LoadedModel instance to be
        constructed in pieces with chained '.' getattr semantics.
        """

        def __init__(self):
            self._model_to_build = LoadedModel()

        def model_future_factory(
            self, caikit_model_future_factory: CaikitModelFutureFactory
        ) -> "LoadedModel.Builder":
            self._model_to_build._caikit_model_future_factory = (
                caikit_model_future_factory
            )
            return self

        def fail_callback(self, callback: Callable) -> "LoadedModel.Builder":
            self._model_to_build._fail_callback = callback
            return self

        def path(self, model_path: str) -> "LoadedModel.Builder":
            self._model_to_build._model_path = model_path
            return self

        def type(self, model_type: str) -> "LoadedModel.Builder":
            self._model_to_build._model_type = model_type
            return self

        def id(self, model_id: str) -> "LoadedModel.Builder":
            self._model_to_build._model_id = model_id
            return self

        def retries(self, retries: int) -> "LoadedModel.Builder":
            self._model_to_build._retries = retries
            return self

        def build(self) -> "LoadedModel":
            error.value_check(
                "<RUN12786023E>",
                self._model_to_build._caikit_model_future_factory
                and self._model_to_build._model_id
                and self._model_to_build._model_type,
                "Cannot build LoadedModel with incomplete required fields."
                + " Future: {}, ID: {}, Type: {}",
                self._model_to_build._caikit_model_future,
                self._model_to_build._model_id,
                self._model_to_build._model_type,
            )
            self._model_to_build._caikit_model_future = (
                self._model_to_build._caikit_model_future_factory()
            )
            return self._model_to_build

    def __init__(self):
        # Use the builder ^^
        self._caikit_model_future_factory: Optional[CaikitModelFutureFactory] = None
        self._caikit_model_future: Optional[CaikitModelFuture] = None
        self._model: Optional[ModuleBase] = None
        self._fail_callback: Optional[Callable] = None
        self._model_id: str = ""
        self._model_path: str = ""
        self._model_type: str = ""
        self._retries: int = 0
        self._size: Optional[int] = None

    def id(self) -> str:
        return self._model_id

    def model(self) -> ModuleBase:
        self.wait()
        return self._model

    def wait(self):
        if self._model is None:
            try:
                self._model = self._caikit_model_future.result()
            except CaikitRuntimeException:
                if self._retries:
                    self._retries -= 1
                    log.debug(
                        "Failed to load %s from %s. Retrying with %d retries left",
                        self.id,
                        self.path,
                        self._retries,
                    )
                    self._caikit_model_future = self._caikit_model_future_factory()
                    # Try waiting again with a fresh load future. This may open
                    # a recursive retry. Once all retries are exhausted, if the
                    # load still fails, the deepest call will invoke the fail
                    # callback and the exception will percolate up from there to
                    # here where it will be raised to the external waiter.
                    self.wait()
                else:
                    if self._fail_callback:
                        self._fail_callback()
                    raise

    def type(self) -> str:
        return self._model_type

    def path(self) -> str:
        return self._model_path

    def size(self) -> int:
        if self._size is None:
            return 0
        return self._size

    def has_size(self) -> bool:
        return self._size is not None and self._size > 0

    # Size is the only mutable member, but only mutable once
    def set_size(self, model_size: int):
        if self._size is None or self._size == 0:
            self._size = model_size
        elif self._size != model_size:
            log.warning(
                "<RUN46815705W>",
                "Attempted to set size of model %s to %s, but it was already %s",
                self.id(),
                model_size,
                self.size(),
            )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Scheme for randomly applying 1+ augmentors, where multiple augmentors may be applied.
"""
# Standard
import random

# First Party
import alog

# Local
from ...exceptions import error_handler
from .base import SchemeBase

log = alog.use_channel("RMULT_AUG_SCHEME")
error = error_handler.get(log)


class RandomMultiSelectionScheme(SchemeBase):
    def __init__(self, preserve_order, selection_probs, augmentors, random_seed=1001):
        """Create a merging augmentor scheme which allows for application of many augmentors
        simultaneously, each of which is controlled by an independent application probability.

        Args:
            preserve_order (bool): Indicates whether or not the contained
                augmentors should always be considered in the order that they
                were provided when they are being applied.
            selection_probs (list(int|float) | tuple(int|float)): Independent
                probability values for applying each augmentor.
            augmentors (list(AugmentorBase) | tuple(AugmentorBase)): Augmentors
                to be applied (in same order as selection_probs).
            random_seed (int): Random seed for controlling shuffling behavior.
        """
        super().__init__(preserve_order, augmentors, random_seed)
        error.type_check("<COR99517020E>", list, tuple, selection_probs=selection_probs)
        error.type_check_all(
            "<COR64536739E>", int, float, selection_probs=selection_probs
        )
        error.value_check(
            "<COR89794223E>",
            len(selection_probs) == len(augmentors),
            "Number of selection probabilties must match the number of augmentors",
        )
        error.value_check(
            "<COR74163144E>",
            all(0 <= prob <= 1 for prob in selection_probs),
            "Selection probabilities must be in the range [0, 1]",
        )
        self._selection_probs = selection_probs

    def _execute(self, obj):
        """Execute the merged scheme by sequentially applying augmentors (in a potentially
        shuffled ordering, based on the value of self.preserve_order).

        Args:
            obj (str | caikit.core.data_model.DataBase): Object to be augmented.
        Returns:
            str | caikit.core.data_model.DataBase: Augmented object of same type
                as input obj.
        """
        output_obj = obj
        for idx in self._current_order:
            selection_prob = self._selection_probs[idx]
            aug = self._augmentors[idx]
            if random.random() <= selection_prob:
                output_obj = aug.augment(output_obj)
        return output_obj
